"""Evaluation related classes and functions."""

import collections
import os
import six

import tensorflow as tf

from opennmt.data import dataset as dataset_lib
from opennmt.utils import misc
from opennmt.utils import scorers as scorers_lib


_SUMMARIES_SCOPE = "metrics"


class EarlyStopping(
    collections.namedtuple("EarlyStopping",
                           ("metric", "min_improvement", "steps"))):
  """Conditions for early stopping."""


class Evaluator(object):
  """Model evaluator."""

  def __init__(self,
               model,
               features_file,
               labels_file,
               batch_size,
               scorers=None,
               save_predictions=False,
               early_stopping=None,
               eval_dir=None):
    """Initializes the evaluator.

    Args:
      model: A :class:`opennmt.models.model.Model` to evaluate.
      features_file: Path to the evaluation features.
      labels_file: Path to the evaluation labels.
      batch_size: The evaluation batch size.
      scorers: A list of scorers, callables taking the path to the reference and
        the hypothesis and return one or more scores.
      save_predictions: Save evaluation predictions to a file. This is ``True``
        when :obj:`external_evaluator` is set.
      early_stopping: An ``EarlyStopping`` instance.
      eval_dir: Directory where predictions can be saved.

    Raises:
      ValueError: If predictions should be saved but the model is not compatible.
      ValueError: If predictions should be saved but :obj:`eval_dir` is ``None``.
      ValueError: If the :obj:`early_stopping` configuration is invalid.
    """
    if scorers is None:
      scorers = []
    if scorers:
      save_predictions = True
    if save_predictions:
      if model.unsupervised:
        raise ValueError("This model does not support saving evaluation predictions")
      if eval_dir is None:
        raise ValueError("Saving evaluation predictions requires eval_dir to be set")
      if not tf.io.gfile.exists(eval_dir):
        tf.io.gfile.makedirs(eval_dir)
    self._model = model
    self._labels_file = labels_file
    self._save_predictions = save_predictions
    self._scorers = scorers
    self._eval_dir = eval_dir
    self._metrics_history = []
    if eval_dir is not None:
      self._summary_writer = tf.summary.create_file_writer(eval_dir)
      summaries = misc.read_summaries(eval_dir)
      for step, values in summaries:
        metrics = misc.extract_prefixed_keys(values, _SUMMARIES_SCOPE + "/")
        self._metrics_history.append((step, metrics))
    else:
      self._summary_writer = tf.summary.create_noop_writer()
    dataset = model.examples_inputter.make_evaluation_dataset(
        features_file,
        labels_file,
        batch_size,
        num_threads=1,
        prefetch_buffer_size=1)

    @dataset_lib.function_on_next(dataset)
    def _eval(next_fn):
      source, target = next_fn()
      outputs, predictions = model(source, labels=target)
      loss = model.compute_loss(outputs, target, training=False)
      return loss, predictions, target

    self._eval = _eval

    self._metrics_name = {"loss", "perplexity"}
    for scorer in self._scorers:
      self._metrics_name.update(scorer.scores_name)
    model_metrics = self._model.get_metrics()
    if model_metrics:
      self._metrics_name.update(set(six.iterkeys(model_metrics)))

    if early_stopping is not None:
      if early_stopping.metric not in self._metrics_name:
        raise ValueError("Invalid early stopping metric '%s', expected one in %s" % (
            early_stopping.metric, str(self._metrics_name)))
      if early_stopping.steps <= 0:
        raise ValueError("Early stopping steps should greater than 0")
    self._early_stopping = early_stopping

  @classmethod
  def from_config(cls, model, config, features_file=None, labels_file=None):
    """Creates an evaluator from the configuration.

    Args:
      model: A :class:`opennmt.models.model.Model` to evaluate.
      config: The global user configuration.
      features_file: Optional input features file to evaluate. If not set, will
        load ``eval_features_file`` from the data configuration.
      labels_file: Optional output labels file to evaluate. If not set, will load
        ``eval_labels_file`` from the data configuration.

    Returns:
      A :class:`opennmt.evaluation.Evaluator` instance.

    Raises:
      ValueError: if one of :obj:`features_file` and :obj:`labels_file` is set
        but not the other.
    """
    if (features_file is None) != (labels_file is None):
      raise ValueError("features_file and labels_file should be both set for evaluation")
    scorers = config["eval"].get("external_evaluators")
    if scorers is not None:
      scorers = scorers_lib.make_scorers(scorers)
    early_stopping_config = config["eval"].get("early_stopping")
    if early_stopping_config is not None:
      early_stopping = EarlyStopping(
          metric=early_stopping_config.get("metric", "loss"),
          min_improvement=early_stopping_config.get("min_improvement", 0),
          steps=early_stopping_config["steps"])
    else:
      early_stopping = None
    return cls(
        model,
        features_file or config["data"]["eval_features_file"],
        labels_file or config["data"].get("eval_labels_file"),
        config["eval"]["batch_size"],
        scorers=scorers,
        save_predictions=config["eval"].get("save_eval_predictions", False),
        early_stopping=early_stopping,
        eval_dir=os.path.join(config["model_dir"], "eval"))

  @property
  def metrics_name(self):
    """The name of the metrics returned by this evaluator."""
    return self._metrics_name

  @property
  def metrics_history(self):
    """The history of metrics result per evaluation step."""
    return self._metrics_history

  @property
  def last_evaluated_step(self):
    """The last training step that was evaluated."""
    if not self._metrics_history:
      return None
    return self._metrics_history[-1][0]

  def _is_higher_better_for_metric(self, metric):
    # Look if the metric is produced by a scorer as they define the scores order.
    for scorer in self._scorers:
      if metric in scorer.scores_name:
        return scorer.higher_is_better()
    # TODO: the condition below is not always true, find a way to set it
    # correctly for Keras metrics.
    return metric not in ("loss", "perplexity")

  def _get_metric_history(self, metric):
    return [
        metrics[metric] for _, metrics in self._metrics_history if metric in metrics]

  def should_stop(self):
    """Returns ``True`` if early stopping conditions are met."""
    if self._early_stopping is None:
      return False
    target_metric = self._early_stopping.metric
    higher_is_better = self._is_higher_better_for_metric(target_metric)
    metrics = self._get_metric_history(target_metric)
    should_stop = early_stop(
        metrics,
        self._early_stopping.steps,
        min_improvement=self._early_stopping.min_improvement,
        higher_is_better=higher_is_better)
    if should_stop:
      tf.get_logger().warning(
          "Evaluation metric '%s' did not improve more than %f in the last %d evaluations",
          target_metric,
          self._early_stopping.min_improvement,
          self._early_stopping.steps)
    return should_stop

  def is_best(self, metric):
    """Returns ``True`` if the latest value of :obj:`metric` is the best so far.

    Args:
      metric: The metric to consider.

    Returns:
      A boolean.
    """
    metric_history = self._get_metric_history(metric)
    if not metric_history:
      return False
    metric_history, latest_value = metric_history[:-1], metric_history[-1]
    if not metric_history:
      return True
    if self._is_higher_better_for_metric(metric):
      return latest_value > max(metric_history)
    else:
      return latest_value < min(metric_history)

  def __call__(self, step):
    """Runs the evaluator.

    Args:
      step: The current training step.

    Returns:
      A dictionary of evaluation metrics.
    """
    tf.get_logger().info("Running evaluation for step %d", step)
    output_file = None
    output_path = None
    if self._save_predictions:
      output_path = os.path.join(self._eval_dir, "predictions.txt.%d" % step)
      output_file = tf.io.gfile.GFile(output_path, "w")

    loss_num = 0
    loss_den = 0
    metrics = self._model.get_metrics()
    for loss, predictions, target in self._eval():  # pylint: disable=no-value-for-parameter
      if isinstance(loss, tuple):
        loss_num += loss[0]
        loss_den += loss[1]
      else:
        loss_num += loss
        loss_den += 1
      if metrics:
        self._model.update_metrics(metrics, predictions, target)
      if output_file is not None:
        predictions = {k:v.numpy() for k, v in six.iteritems(predictions)}
        for prediction in misc.extract_batches(predictions):
          self._model.print_prediction(prediction, stream=output_file)
    if loss_den == 0:
      raise RuntimeError("No examples were evaluated")
    loss = loss_num / loss_den

    results = dict(loss=loss, perplexity=tf.math.exp(loss))
    if metrics:
      for name, metric in six.iteritems(metrics):
        results[name] = metric.result()
    if self._save_predictions:
      tf.get_logger().info("Evaluation predictions saved to %s", output_path)
      output_file.close()
      for scorer in self._scorers:
        score = scorer(self._labels_file, output_path)
        if isinstance(score, dict):
          results.update(score)
        else:
          results[scorer.name] = score

    return self._record_results(step, results)

  def _record_results(self, step, results):
    for name, value in six.iteritems(results):
      if isinstance(value, tf.Tensor):
        results[name] = value.numpy()
    # Clear history for steps that are greater than step.
    while self._metrics_history and self._metrics_history[-1][0] > step:
      self._metrics_history.pop()
    self._metrics_history.append((step, dict(results)))
    tf.get_logger().info(
        "Evaluation result for step %d: %s",
        step,
        " ; ".join("%s = %f" % (k, v) for k, v in six.iteritems(results)))
    with self._summary_writer.as_default():
      for key, value in six.iteritems(results):
        tf.summary.scalar("%s/%s" % (_SUMMARIES_SCOPE, key), value, step=step)
      self._summary_writer.flush()
    return results


def early_stop(metrics, steps, min_improvement=0, higher_is_better=False):
  """Early stopping condition.

  Args:
    metrics: A list of metric values.
    steps: Consider the improvement over this many steps.
    min_improvement: Continue if the metric improved less than this value:
    higher_is_better: Whether a higher value is better for this metric.

  Returns:
    A boolean.
  """
  if len(metrics) < steps + 1:
    return False

  def _did_improve(ref, new):
    # Returns True if new is improving on ref.
    if higher_is_better:
      return new > ref + min_improvement
    else:
      return new < ref - min_improvement

  ref_metric = metrics[-steps - 1]
  for metric in metrics[-steps:]:
    if _did_improve(ref_metric, metric):
      return False
  return True



"""Various tensor manipulation functions."""

import tensorflow as tf


def roll_sequence(tensor, offsets):
  """Shifts sequences by an offset.

  Args:
    tensor: A ``tf.Tensor`` of shape :math:`[B, T, ...]`.
    offsets : The offset of each sequence of shape :math:`[B]`.

  Returns:
    A ``tf.Tensor`` with the same shape as :obj:`tensor` and with sequences
    shifted by :obj:`offsets`.
  """
  batch_size = tf.shape(tensor)[0]
  time = tf.shape(tensor)[1]
  cols, rows = tf.meshgrid(tf.range(time), tf.range(batch_size))
  cols -= tf.expand_dims(offsets, 1)
  cols %= time
  indices = tf.stack([rows, cols], axis=-1)
  return tf.gather_nd(tensor, indices)



"""Module defining reusable and model specific layers."""

from opennmt.layers.bridge import Bridge
from opennmt.layers.bridge import CopyBridge
from opennmt.layers.bridge import DenseBridge
from opennmt.layers.bridge import ZeroBridge

from opennmt.layers.common import Dense
from opennmt.layers.common import LayerNorm
from opennmt.layers.common import LayerWrapper
from opennmt.layers.common import dropout
from opennmt.layers.common import gelu

from opennmt.layers.position import PositionEmbedder
from opennmt.layers.position import PositionEncoder
from opennmt.layers.position import SinusoidalPositionEncoder

from opennmt.layers.reducer import ConcatReducer
from opennmt.layers.reducer import JoinReducer
from opennmt.layers.reducer import MultiplyReducer
from opennmt.layers.reducer import Reducer
from opennmt.layers.reducer import SumReducer

from opennmt.layers.rnn import LSTM
from opennmt.layers.rnn import RNN
from opennmt.layers.rnn import RNNCellWrapper
from opennmt.layers.rnn import make_rnn_cell

from opennmt.layers.transformer import FeedForwardNetwork
from opennmt.layers.transformer import MultiHeadAttention
from opennmt.layers.transformer import SelfAttentionDecoderLayer
from opennmt.layers.transformer import SelfAttentionEncoderLayer
from opennmt.layers.transformer import TransformerLayerWrapper
from opennmt.layers.transformer import combine_heads
from opennmt.layers.transformer import future_mask
from opennmt.layers.transformer import split_heads



"""Define the OpenNMT tokenizer."""

import os
import copy
import six
import yaml

import tensorflow as tf

import pyonmttok

from opennmt.tokenizers.tokenizer import Tokenizer


def _make_config_asset_file(config, asset_path):
  asset_config = copy.deepcopy(config)
  for key, value in six.iteritems(asset_config):
    # Only keep the basename for files (that should also be registered as assets).
    if isinstance(value, six.string_types) and tf.io.gfile.exists(value):
      asset_config[key] = os.path.basename(value)
  with tf.io.gfile.GFile(asset_path, "w") as asset_file:
    yaml.dump(asset_config, stream=asset_file, default_flow_style=False)


class OpenNMTTokenizer(Tokenizer):
  """Uses the OpenNMT tokenizer."""

  def __init__(self, **kwargs):
    self._config = copy.deepcopy(kwargs)
    mode = "conservative"
    if "mode" in kwargs:
      mode = kwargs["mode"]
      del kwargs["mode"]
    self._tokenizer = pyonmttok.Tokenizer(mode, **kwargs)

  def export_assets(self, asset_dir, asset_prefix=""):
    asset_name = "%stokenizer_config.yml" % asset_prefix
    asset_path = os.path.join(asset_dir, asset_name)
    _make_config_asset_file(self._config, asset_path)
    assets = {}
    assets[asset_name] = asset_path
    for key, value in six.iteritems(self._config):
      if key.endswith("path"):
        assets[os.path.basename(value)] = value
    return assets

  def _tokenize_string(self, text):
    text = tf.compat.as_bytes(text)
    tokens, _ = self._tokenizer.tokenize(text)
    return [tf.compat.as_text(token) for token in tokens]

  def _detokenize_string(self, tokens):
    tokens = [tf.compat.as_bytes(token) for token in tokens]
    return tf.compat.as_text(self._tokenizer.detokenize(tokens))



"""Main script."""

import argparse
import logging
import os
import six

import tensorflow as tf

from opennmt import __version__
from opennmt.models import catalog
from opennmt.runner import Runner
from opennmt.config import load_model, load_config
from opennmt.utils.misc import classes_in_module


_PYTHON_TO_TENSORFLOW_LOGGING_LEVEL = {
    logging.CRITICAL: 3,
    logging.ERROR: 2,
    logging.WARNING: 1,
    logging.INFO: 0,
    logging.DEBUG: 0,
    logging.NOTSET: 0,
}

def _set_log_level(log_level):
  tf.get_logger().setLevel(log_level)
  os.environ["TF_CPP_MIN_LOG_LEVEL"] = str(_PYTHON_TO_TENSORFLOW_LOGGING_LEVEL[log_level])

def _prefix_paths(prefix, paths):
  """Recursively prefix paths.

  Args:
    prefix: The prefix to apply.
    data: A dict of relative paths.

  Returns:
    The updated dict.
  """
  if isinstance(paths, dict):
    for key, path in six.iteritems(paths):
      paths[key] = _prefix_paths(prefix, path)
    return paths
  elif isinstance(paths, list):
    for i, path in enumerate(paths):
      paths[i] = _prefix_paths(prefix, path)
    return paths
  else:
    path = paths
    new_path = os.path.join(prefix, path)
    if tf.io.gfile.exists(new_path):
      return new_path
    else:
      return path

def main():
  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument("-v", "--version", action="version", version="OpenNMT-tf %s" % __version__)
  parser.add_argument("--config", required=True, nargs="+",
                      help="List of configuration files.")
  parser.add_argument("--auto_config", default=False, action="store_true",
                      help="Enable automatic configuration values.")
  parser.add_argument("--model_type", default="",
                      choices=list(classes_in_module(catalog, public_only=True)),
                      help="Model type from the catalog.")
  parser.add_argument("--model", default="",
                      help="Custom model configuration file.")
  parser.add_argument("--run_dir", default="",
                      help="If set, model_dir will be created relative to this location.")
  parser.add_argument("--data_dir", default="",
                      help="If set, data files are expected to be relative to this location.")
  parser.add_argument("--checkpoint_path", default=None,
                      help=("Specific checkpoint or model directory to load "
                            "(when a directory is set, the latest checkpoint is used)."))
  parser.add_argument("--log_level", default="INFO",
                      choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
                      help="Logs verbosity.")
  parser.add_argument("--seed", type=int, default=None,
                      help="Random seed.")
  parser.add_argument("--gpu_allow_growth", default=False, action="store_true",
                      help="Allocate GPU memory dynamically.")
  parser.add_argument("--intra_op_parallelism_threads", type=int, default=0,
                      help=("Number of intra op threads (0 means the system picks "
                            "an appropriate number)."))
  parser.add_argument("--inter_op_parallelism_threads", type=int, default=0,
                      help=("Number of inter op threads (0 means the system picks "
                            "an appropriate number)."))
  parser.add_argument("--mixed_precision", default=False, action="store_true",
                      help="Enable mixed precision.")

  subparsers = parser.add_subparsers(help="Run type.", dest="run")
  parser_train = subparsers.add_parser("train", help="Training.")
  parser_train.add_argument(
      "--with_eval", default=False, action="store_true",
      help="Enable automatic evaluation.")
  parser_train.add_argument(
      "--num_gpus", type=int, default=1,
      help="Number of GPUs to use for in-graph replication.")

  parser_eval = subparsers.add_parser("eval", help="Evaluation.")
  parser_eval.add_argument(
      "--features_file", nargs="+", default=None,
      help="Input features files.")
  parser_eval.add_argument(
      "--labels_file", default=None,
      help="Output labels files.")

  parser_infer = subparsers.add_parser("infer", help="Inference.")
  parser_infer.add_argument(
      "--features_file", nargs="+", required=True,
      help="Run inference on this file.")
  parser_infer.add_argument(
      "--predictions_file", default="",
      help=("File used to save predictions. If not set, predictions are printed "
            "on the standard output."))
  parser_infer.add_argument(
      "--log_prediction_time", default=False, action="store_true",
      help="Logs some prediction time metrics.")

  parser_export = subparsers.add_parser("export", help="Model export.")
  parser_export.add_argument(
      "--export_dir", required=True,
      help="The directory of the exported model.")

  parser_score = subparsers.add_parser("score", help="Scoring.")
  parser_score.add_argument("--features_file", nargs="+", required=True,
                            help="Features file.")
  parser_score.add_argument("--predictions_file", default=None,
                            help="Predictions to score.")

  parser_average_checkpoints = subparsers.add_parser(
      "average_checkpoints", help="Checkpoint averaging.")
  parser_average_checkpoints.add_argument(
      "--output_dir", required=True,
      help="The output directory for the averaged checkpoint.")
  parser_average_checkpoints.add_argument(
      "--max_count", type=int, default=8,
      help="The maximal number of checkpoints to average.")

  parser_update_vocab = subparsers.add_parser(
      "update_vocab", help="Update model vocabularies in checkpoint.")
  parser_update_vocab.add_argument(
      "--output_dir", required=True,
      help="The output directory for the updated checkpoint.")
  parser_update_vocab.add_argument(
      "--src_vocab", default=None,
      help="Path to the new source vocabulary.")
  parser_update_vocab.add_argument(
      "--tgt_vocab", default=None,
      help="Path to the new target vocabulary.")

  args = parser.parse_args()
  if hasattr(args, "features_file") and args.features_file and len(args.features_file) == 1:
    args.features_file = args.features_file[0]

  _set_log_level(getattr(logging, args.log_level))
  tf.config.threading.set_intra_op_parallelism_threads(args.intra_op_parallelism_threads)
  tf.config.threading.set_inter_op_parallelism_threads(args.inter_op_parallelism_threads)
  if args.gpu_allow_growth:
    for device in tf.config.experimental.list_physical_devices(device_type="GPU"):
      tf.config.experimental.set_memory_growth(device, enable=True)

  # Load and merge run configurations.
  config = load_config(args.config)
  if args.run_dir:
    config["model_dir"] = os.path.join(args.run_dir, config["model_dir"])
  if args.data_dir:
    config["data"] = _prefix_paths(args.data_dir, config["data"])

  if not tf.io.gfile.exists(config["model_dir"]):
    tf.get_logger().info("Creating model directory %s", config["model_dir"])
    tf.io.gfile.makedirs(config["model_dir"])

  model = load_model(
      config["model_dir"],
      model_file=args.model,
      model_name=args.model_type)
  runner = Runner(
      model,
      config,
      auto_config=args.auto_config,
      mixed_precision=args.mixed_precision,
      seed=args.seed)

  if args.run == "train":
    runner.train(
        num_devices=args.num_gpus,
        with_eval=args.with_eval,
        checkpoint_path=args.checkpoint_path)
  elif args.run == "eval":
    metrics = runner.evaluate(
        checkpoint_path=args.checkpoint_path,
        features_file=args.features_file,
        labels_file=args.labels_file)
    print(metrics)
  elif args.run == "infer":
    runner.infer(
        args.features_file,
        predictions_file=args.predictions_file,
        checkpoint_path=args.checkpoint_path,
        log_time=args.log_prediction_time)
  elif args.run == "export":
    runner.export(
        args.export_dir,
        checkpoint_path=args.checkpoint_path)
  elif args.run == "score":
    runner.score(
        args.features_file,
        args.predictions_file,
        checkpoint_path=args.checkpoint_path)
  elif args.run == "average_checkpoints":
    runner.average_checkpoints(args.output_dir, max_count=args.max_count)
  elif args.run == "update_vocab":
    runner.update_vocab(
        args.output_dir,
        src_vocab=args.src_vocab,
        tgt_vocab=args.tgt_vocab)


if __name__ == "__main__":
  main()



"""Define bridges: logic of passing the encoder state to the decoder."""

import abc
import six

import tensorflow as tf


def assert_state_is_compatible(expected_state, state):
  """Asserts that states are compatible.

  Args:
    expected_state: The reference state.
    state: The state that must be compatible with :obj:`expected_state`.

  Raises:
    ValueError: if the states are incompatible.
  """
  # Check structure compatibility.
  tf.nest.assert_same_structure(expected_state, state)

  # Check shape compatibility.
  expected_state_flat = tf.nest.flatten(expected_state)
  state_flat = tf.nest.flatten(state)

  for x, y in zip(expected_state_flat, state_flat):
    if tf.is_tensor(x):
      expected_depth = x.shape[-1]
      depth = y.shape[-1]
      if depth != expected_depth:
        raise ValueError("Tensor in state has shape %s which is incompatible "
                         "with the target shape %s" % (y.shape, x.shape))


@six.add_metaclass(abc.ABCMeta)
class Bridge(tf.keras.layers.Layer):
  """Base class for bridges."""

  def __call__(self, encoder_state, decoder_zero_state):  # pylint: disable=arguments-differ
    """Returns the initial decoder state.

    Args:
      encoder_state: The encoder state.
      decoder_zero_state: The default decoder state.

    Returns:
      The decoder initial state.
    """
    return super(Bridge, self).__call__([encoder_state, decoder_zero_state])

  @abc.abstractmethod
  def call(self, states):  # pylint: disable=arguments-differ
    raise NotImplementedError()


class CopyBridge(Bridge):
  """A bridge that passes the encoder state as is."""

  def call(self, states):
    encoder_state, decoder_state = states
    assert_state_is_compatible(encoder_state, decoder_state)
    flat_encoder_state = tf.nest.flatten(encoder_state)
    return tf.nest.pack_sequence_as(decoder_state, flat_encoder_state)


class ZeroBridge(Bridge):
  """A bridge that does not pass information from the encoder."""

  def call(self, states):
    # Simply return the default decoder state.
    return states[1]


class DenseBridge(Bridge):
  """A bridge that applies a parameterized linear transformation from the
  encoder state to the decoder state size.
  """

  def __init__(self, activation=None):
    """Initializes the bridge.

    Args:
      activation: Activation function (a callable).
        Set it to ``None`` to maintain a linear activation.
    """
    super(DenseBridge, self).__init__()
    self.activation = activation
    self.decoder_state_sizes = None
    self.linear = None

  def build(self, input_shape):
    decoder_shape = input_shape[1]
    self.decoder_state_sizes = [
        shape[-1] for shape in tf.nest.flatten(decoder_shape)]
    self.linear = tf.keras.layers.Dense(
        sum(self.decoder_state_sizes), activation=self.activation)

  def call(self, states):
    encoder_state, decoder_state = states
    encoder_state_flat = tf.nest.flatten(encoder_state)
    encoder_state_single = tf.concat(encoder_state_flat, 1)
    transformed = self.linear(encoder_state_single)
    splitted = tf.split(transformed, self.decoder_state_sizes, axis=1)
    return tf.nest.pack_sequence_as(decoder_state, splitted)



"""Define RNN-based encoders."""

import tensorflow as tf
import tensorflow_addons as tfa

from opennmt.encoders.encoder import Encoder, SequentialEncoder
from opennmt.layers.reducer import ConcatReducer, JoinReducer, pad_in_time
from opennmt.layers import common
from opennmt.layers import rnn


class _RNNEncoderBase(Encoder):
  """Base class for RNN encoders."""

  def __init__(self, rnn_layer, **kwargs):
    """Initializes the encoder.

    Args:
      rnn_layer: The RNN layer used to encode the inputs.
      **kwargs: Additional layer arguments.
    """
    super(_RNNEncoderBase, self).__init__(**kwargs)
    self.rnn = rnn_layer

  def call(self, inputs, sequence_length=None, training=None):
    mask = self.build_mask(inputs, sequence_length=sequence_length)
    outputs, states = self.rnn(inputs, mask=mask, training=training)
    return outputs, states, sequence_length


class RNNEncoder(_RNNEncoderBase):
  """A RNN sequence encoder."""

  def __init__(self,
               num_layers,
               num_units,
               bidirectional=False,
               residual_connections=False,
               dropout=0.3,
               reducer=ConcatReducer(),
               cell_class=None,
               **kwargs):
    """Initializes the parameters of the encoder.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer.
      bidirectional: Use a bidirectional RNN.
      residual_connections: If ``True``, each layer input will be added to its
        output.
      dropout: The probability to drop units in each layer output.
      reducer: A :class:`opennmt.layers.Reducer` instance to merge
        bidirectional state and outputs.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell. Defaults to a LSTM cell.
      **kwargs: Additional layer arguments.
    """
    cell = rnn.make_rnn_cell(
        num_layers,
        num_units,
        dropout=dropout,
        residual_connections=residual_connections,
        cell_class=cell_class)
    rnn_layer = rnn.RNN(cell, bidirectional=bidirectional, reducer=reducer)
    super(RNNEncoder, self).__init__(rnn_layer, **kwargs)


class LSTMEncoder(_RNNEncoderBase):
  """A LSTM sequence encoder.

  See Also:
    :class:`opennmt.layers.LSTM` for differences between this encoder and
    :class:`opennmt.encoders.RNNEncoder` with a `LSTMCell`.
  """

  def __init__(self,
               num_layers,
               num_units,
               bidirectional=False,
               residual_connections=False,
               dropout=0.3,
               reducer=ConcatReducer(),
               **kwargs):
    """Initializes the parameters of the encoder.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer output.
      bidirectional: Make each LSTM layer bidirectional.
      residual_connections: If ``True``, each layer input will be added to its
        output.
      dropout: The probability to drop units in each layer output.
      reducer: A :class:`opennmt.layers.Reducer` instance to merge
        bidirectional state and outputs.
      **kwargs: Additional layer arguments.
    """
    lstm_layer = rnn.LSTM(
        num_layers,
        num_units,
        bidirectional=bidirectional,
        reducer=reducer,
        dropout=dropout,
        residual_connections=residual_connections)
    super(LSTMEncoder, self).__init__(lstm_layer, **kwargs)


class GNMTEncoder(SequentialEncoder):
  """The RNN encoder used in GNMT as described in
  https://arxiv.org/abs/1609.08144.
  """

  def __init__(self, num_layers, num_units, dropout=0.3):
    """Initializes the parameters of the encoder.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer.
      dropout: The probability to drop units in each layer output.

    Raises:
      ValueError: if :obj:`num_layers` < 2.
    """
    if num_layers < 2:
      raise ValueError("GNMTEncoder requires at least 2 layers")
    bidirectional = LSTMEncoder(
        1,
        num_units,
        bidirectional=True,
        dropout=dropout)
    unidirectional = LSTMEncoder(
        num_layers - 1,
        num_units,
        dropout=dropout,
        residual_connections=True)
    super(GNMTEncoder, self).__init__([bidirectional, unidirectional])


class RNMTPlusEncoder(SequentialEncoder):
  """The RNMT+ encoder described in https://arxiv.org/abs/1804.09849."""

  def __init__(self,
               num_layers=6,
               num_units=1024,
               cell_class=None,
               dropout=0.3):
    """Initializes the parameters of the encoder.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each RNN layer and the final output.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell. Defaults to a layer normalized LSTM cell.
      dropout: The probability to drop units in each layer output.
    """
    if cell_class is None:
      cell_class = tfa.rnn.LayerNormLSTMCell
    layers = [
        RNNEncoder(
            1,
            num_units,
            bidirectional=True,
            dropout=0.0,
            cell_class=cell_class)
        for _ in range(num_layers)]
    layers = [
        common.LayerWrapper(layer, output_dropout=dropout, residual_connection=True)
        for layer in layers]
    super(RNMTPlusEncoder, self).__init__(layers)
    self.dropout = dropout
    self.projection = tf.keras.layers.Dense(num_units)

  def call(self, inputs, sequence_length=None, training=None):
    inputs = common.dropout(inputs, self.dropout, training=training)
    outputs, state, sequence_length = super(RNMTPlusEncoder, self).call(
        inputs, sequence_length=sequence_length, training=training)
    projected = self.projection(outputs)
    return (projected, state, sequence_length)


class PyramidalRNNEncoder(Encoder):
  """An encoder that reduces the time dimension after each bidirectional layer."""

  def __init__(self,
               num_layers,
               num_units,
               reduction_factor=2,
               cell_class=None,
               dropout=0.3):
    """Initializes the parameters of the encoder.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer.
      reduction_factor: The time reduction factor.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell. Defaults to a LSTM cell.
      dropout: The probability to drop units in each layer output.
    """
    super(PyramidalRNNEncoder, self).__init__()
    self.reduction_factor = reduction_factor
    self.state_reducer = JoinReducer()
    self.layers = [
        RNNEncoder(
            1,
            num_units // 2,
            bidirectional=True,
            reducer=ConcatReducer(),
            cell_class=cell_class,
            dropout=dropout)
        for _ in range(num_layers)]

  def call(self, inputs, sequence_length=None, training=None):
    encoder_state = []

    for layer_index, layer in enumerate(self.layers):
      input_depth = inputs.shape[-1]

      if layer_index == 0:
        # For the first input, make the number of timesteps a multiple of the
        # total reduction factor.
        total_reduction_factor = pow(self.reduction_factor, len(self.layers) - 1)

        current_length = tf.shape(inputs)[1]
        factor = tf.cast(current_length, tf.float32) / total_reduction_factor
        new_length = tf.cast(tf.math.ceil(factor), tf.int32) * total_reduction_factor
        inputs = pad_in_time(inputs, new_length - current_length)

        # Lengths should not be smaller than the total reduction factor.
        sequence_length = tf.maximum(sequence_length, total_reduction_factor)
      else:
        # In other cases, reduce the time dimension.
        inputs = tf.reshape(
            inputs,
            [tf.shape(inputs)[0], -1, input_depth * self.reduction_factor])
        if sequence_length is not None:
          sequence_length //= self.reduction_factor

      outputs, state, sequence_length = layer(
          inputs,
          sequence_length=sequence_length,
          training=training)

      encoder_state.append(state)
      inputs = outputs

    return (
        outputs,
        self.state_reducer(encoder_state),
        sequence_length)



"""Module defining encoders."""

from opennmt.encoders.conv_encoder import ConvEncoder

from opennmt.encoders.encoder import Encoder
from opennmt.encoders.encoder import ParallelEncoder
from opennmt.encoders.encoder import SequentialEncoder

from opennmt.encoders.mean_encoder import MeanEncoder

from opennmt.encoders.rnn_encoder import GNMTEncoder
from opennmt.encoders.rnn_encoder import LSTMEncoder
from opennmt.encoders.rnn_encoder import PyramidalRNNEncoder
from opennmt.encoders.rnn_encoder import RNMTPlusEncoder
from opennmt.encoders.rnn_encoder import RNNEncoder

from opennmt.encoders.self_attention_encoder import SelfAttentionEncoder



# -*- coding: utf-8 -*-

"""Text manipulation."""

import tensorflow as tf


def tokens_to_chars(tokens):
  """Splits tokens into unicode characters.

  Args:
    tokens: A string ``tf.Tensor`` of shape :math:`[T]`.

  Returns:
    The characters as a 2D string ``tf.RaggedTensor``.
  """
  return tf.strings.unicode_split(tokens, "UTF-8")

def tokens_to_words(tokens, subword_token="￭", is_spacer=None):
  """Converts a sequence of tokens to a sequence of words.

  For example, if a BPE tokenization produces this sequence:

      ["He@@", "llo", "W@@", "orld", "@@!"]

  this function will return the tensor:

      [["He@@", "llo", ""], ["W@@", "orld", "@@!"]]

  Args:
    tokens: A 1D string ``tf.Tensor``.
    subword_token: The special token used by the subword tokenizer.
    is_spacer: Whether :obj:`subword_token` is used as a spacer (as in
      SentencePiece) or a joiner (as in BPE). If ``None``, will infer
      directly from :obj:`subword_token`.

  Returns:
    The words as a 2D string ``tf.RaggedTensor``.
  """
  if is_spacer is None:
    is_spacer = subword_token == "▁"
  if is_spacer:
    # First token implicitly starts with a spacer.
    left_and_single = tf.logical_or(
        tf.strings.regex_full_match(tokens, "%s.*" % subword_token),
        tf.one_hot(0, tf.shape(tokens)[0], on_value=True, off_value=False))
    right = tf.strings.regex_full_match(tokens, ".+%s" % subword_token)
    word_start = tf.logical_or(tf.roll(right, shift=1, axis=0), left_and_single)
  else:
    right = tf.strings.regex_full_match(tokens, ".*%s" % subword_token)
    left = tf.strings.regex_full_match(tokens, "%s.*" % subword_token)
    subword = tf.logical_or(tf.roll(right, shift=1, axis=0), left)
    word_start = tf.logical_not(subword)
  start_indices = tf.squeeze(tf.where(word_start), -1)
  return tf.RaggedTensor.from_row_starts(tokens, start_indices)

def alignment_matrix_from_pharaoh(alignment_line,
                                  source_length,
                                  target_length,
                                  dtype=tf.float32):
  """Parse Pharaoh alignments into an alignment matrix.

  Args:
    alignment_line: A string ``tf.Tensor`` in the Pharaoh format.
    source_length: The length of the source sentence, without special symbols.
    target_length: The length of the target sentence, without special symbols.
    dtype: The output matrix dtype. Defaults to ``tf.float32`` for convenience
      when computing the guided alignment loss.

  Returns:
    The alignment matrix as a 2-D ``tf.Tensor`` of type :obj:`dtype` and shape
    ``[target_length, source_length]``, where ``[i, j] = 1`` if the ``i`` th
    target word is aligned with the ``j`` th source word.
  """
  align_pairs_str = tf.strings.split([alignment_line]).values
  align_pairs_flat_str = tf.strings.split(align_pairs_str, sep="-").values
  align_pairs_flat = tf.strings.to_number(align_pairs_flat_str, out_type=tf.int64)
  sparse_indices = tf.reshape(align_pairs_flat, [-1, 2])
  sparse_values = tf.ones([tf.shape(sparse_indices)[0]], dtype=dtype)
  source_length = tf.cast(source_length, tf.int64)
  target_length = tf.cast(target_length, tf.int64)
  alignment_matrix_sparse = tf.sparse.SparseTensor(
      sparse_indices, sparse_values, [source_length, target_length])
  alignment_matrix = tf.sparse.to_dense(alignment_matrix_sparse, validate_indices=False)
  return tf.transpose(alignment_matrix)



"""Defines common layers."""

import tensorflow as tf
import numpy as np

from opennmt.utils.misc import shape_list


def dropout(x, rate, training=None):
  """Simple dropout layer."""
  if not training or rate == 0:
    return x
  return tf.nn.dropout(x, rate)

def gelu(x):
  """Gaussian Error Linear Unit activation function described in
  https://arxiv.org/abs/1606.08415.
  """
  return 0.5 * x * (1 + tf.tanh(np.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3))))


class Dense(tf.keras.layers.Dense):
  """Small ``tf.keras.layers.Dense`` extension to possibly reuse an existing weight
  matrix.
  """

  def __init__(self, units, weight=None, transpose=False, **kwargs):
    """Initializes the layer.

    Args:
      unit: Positive integer, dimensionality of the output space.
      weight: The weight to reuse.
      transpose: Whether :obj:`weight` should be transposed or not.
      kwargs: Additional layers arguments.
    """
    super(Dense, self).__init__(units, **kwargs)
    self.weight = weight
    self.transpose = transpose

  def add_weight(self, name, *args, **kwargs):  # pylint: disable=arguments-differ
    if self.weight is not None and name == "kernel":
      return self.weight
    return super(Dense, self).add_weight(name, *args, **kwargs)

  def call(self, inputs):
    shape = shape_list(inputs)
    rank = len(shape)
    if rank > 2:
      inputs = tf.reshape(inputs, [-1, shape[-1]])
    outputs = tf.matmul(inputs, self.kernel, transpose_b=self.transpose)
    if self.use_bias:
      outputs = tf.nn.bias_add(outputs, self.bias)
    if self.activation is not None:
      outputs = self.activation(outputs)  # pylint: disable=not-callable
    if rank > 2:
      outputs = tf.reshape(outputs, shape[:-1] + [self.units])
    return outputs

  def map_v1_weights(self, weights):
    m = [(self.kernel, weights["kernel"])]
    if self.use_bias:
      m.append((self.bias, weights["bias"]))
    return m


class LayerNorm(tf.keras.layers.Layer):
  """Layer normalization."""

  def __init__(self, epsilon=1e-6, **kwargs):
    """Initializes this layer.

    Args:
      epsilon: The epsilon value to use.
      kwargs: Additional layer arguments.
    """
    super(LayerNorm, self).__init__(**kwargs)
    self.epsilon = epsilon

  def build(self, input_shape):
    """Creates the variables."""
    depth = input_shape[-1]
    self.beta = self.add_weight(
        "beta", [depth], initializer=tf.keras.initializers.Constant(0))
    self.gamma = self.add_weight(
        "gamma", [depth], initializer=tf.keras.initializers.Constant(1))
    super(LayerNorm, self).build(input_shape)

  def call(self, x):  # pylint: disable=arguments-differ
    """Normalizes :obj:`x`."""
    mean = tf.reduce_mean(x, axis=[-1], keepdims=True)
    variance = tf.reduce_mean(tf.square(x - mean), axis=[-1], keepdims=True)
    norm_x = (x - mean) * tf.math.rsqrt(variance + self.epsilon)
    return norm_x * self.gamma + self.beta

  def map_v1_weights(self, weights):
    return [
        (self.beta, weights["beta"]),
        (self.gamma, weights["gamma"])
    ]


class LayerWrapper(tf.keras.layers.Layer):
  """Layer wrapper for input/output normalization, input/output dropout and
  residual connection.
  """

  def __init__(self,
               layer,
               normalize_input=False,
               normalize_output=False,
               input_dropout=0,
               output_dropout=0,
               residual_connection=False,
               **kwargs):
    """Initializes the layer.

    Args:
      layer: The layer to wrap.
      normalize_input: Apply layer normalization on the input.
      normalize_output: Apply layer normalization on the output.
      input_dropout: The probability to drop units in the layer input.
      output_dropout: The probability to drop units in the layer output.
      residual_connection: Add the inputs to layer outputs (if their shape are
        compatible).
      kwargs: Additional layer arguments.
    """
    super(LayerWrapper, self).__init__(**kwargs)
    self.layer = layer
    self.input_layer_norm = LayerNorm() if normalize_input else None
    self.output_layer_norm = LayerNorm() if normalize_output else None
    self.input_dropout = input_dropout
    self.output_dropout = output_dropout
    self.residual_connection = residual_connection

  def call(self, inputs, *args, **kwargs):  # pylint: disable=arguments-differ
    """Runs the wrapper."""
    training = kwargs.get("training")
    x = inputs
    if self.input_layer_norm is not None:
      x = self.input_layer_norm(x)  # pylint: disable=not-callable
    x = dropout(x, self.input_dropout, training=training)

    all_outputs = self.layer(x, *args, **kwargs)
    if isinstance(all_outputs, tuple):
      outputs = all_outputs[0]
      extra_outputs = list(all_outputs)[1:]
    else:
      outputs = all_outputs
      extra_outputs = None

    outputs = dropout(outputs, self.output_dropout, training=training)
    if self.residual_connection and outputs.shape[-1] == inputs.shape[-1]:
      outputs += inputs
    if self.output_layer_norm is not None:
      outputs = self.output_layer_norm(outputs)  # pylint: disable=not-callable

    if extra_outputs:
      return tuple([outputs] + extra_outputs)
    return outputs

  # The wrapper should be serializable to be used in tf.keras.layers.Bidirectional.

  def get_config(self):
    """Returns the layer wrapper configuration."""
    config = {
        "layer": tf.keras.layers.serialize(self.layer),
        "normalize_input": self.input_layer_norm is not None,
        "normalize_output": self.output_layer_norm is not None,
        "input_dropout": self.input_dropout,
        "output_dropout": self.output_dropout,
        "residual_connection": self.residual_connection
    }
    base_config = super(LayerWrapper, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  @classmethod
  def from_config(cls, config):
    """Creates a layer wrapper from its configuration."""
    layer = tf.keras.layers.deserialize(config.pop("layer"))
    return cls(layer, **config)



"""RNN functions and classes for TensorFlow 2.0."""

import tensorflow as tf

from opennmt.layers import common
from opennmt.layers import reducer as reducer_lib


def _register_keras_custom_object(cls):
  tf.keras.utils.get_custom_objects()[cls.__name__] = cls
  return cls


@_register_keras_custom_object
class RNNCellWrapper(common.LayerWrapper):
  """A wrapper for RNN cells."""

  def __init__(self,
               cell,
               input_dropout=0,
               output_dropout=0,
               residual_connection=False,
               **kwargs):
    """Initializes the wrapper.

    Args:
      cell: The cell to wrap.
      input_dropout: The probability to drop units in the cell input.
      output_dropout: The probability to drop units in the cell output.
      residual_connection: Add the inputs to cell outputs (if their shape are
        compatible).
      kwargs: Additional layer arguments.
    """
    super(RNNCellWrapper, self).__init__(
        cell,
        input_dropout=input_dropout,
        output_dropout=output_dropout,
        residual_connection=residual_connection,
        **kwargs)
    self.cell = cell

  @property
  def state_size(self):
    """The cell state size."""
    return self.cell.state_size

  @property
  def output_size(self):
    """The cell output size."""
    return self.cell.output_size

  def get_initial_state(self, inputs=None, batch_size=None, dtype=None):
    """Returns the initial cell state."""
    return self.cell.get_initial_state(
        inputs=inputs, batch_size=batch_size, dtype=dtype)


class _StackedRNNCells(tf.keras.layers.StackedRNNCells):

  # To pass the training flag to the cell, tf.keras.layers.RNN checks that the
  # cell call method explicitly takes the "training" argument, which
  # tf.keras.layers.StackedRNNCells do not.
  # TODO: remove this when this change is released:
  # https://github.com/tensorflow/tensorflow/commit/df2b252fa380994cd9236cc56b06557bcf12a9d3
  def call(self, inputs, states, constants=None, training=None, **kwargs):
    kwargs["training"] = training
    return super(_StackedRNNCells, self).call(inputs, states, constants=constants, **kwargs)


def make_rnn_cell(num_layers,
                  num_units,
                  dropout=0,
                  residual_connections=False,
                  cell_class=None,
                  **kwargs):
  """Convenience function to build a multi-layer RNN cell.

  Args:
    num_layers: The number of layers.
    num_units: The number of units in each layer.
    dropout: The probability to drop units in each layer output.
    residual_connections: If ``True``, each layer input will be added to its output.
    cell_class: The inner cell class or a callable taking :obj:`num_units` as
      argument and returning a cell. Defaults to a LSTM cell.
    kwargs: Additional arguments passed to the cell constructor.

  Returns:
    A ``tf.keras.layers.StackedRNNCells`` instance.

  See Also:
    :class:`opennmt.layers.RNNCellWrapper`
  """
  if cell_class is None:
    cell_class = tf.keras.layers.LSTMCell
  cells = []
  for _ in range(num_layers):
    cell = cell_class(num_units, **kwargs)
    if dropout > 0 or residual_connections:
      cell = RNNCellWrapper(
          cell, output_dropout=dropout, residual_connection=residual_connections)
    cells.append(cell)
  return _StackedRNNCells(cells)


class _RNNWrapper(tf.keras.layers.Layer):
  """Extend a RNN layer to possibly make it bidirectional and format its outputs."""

  def __init__(self, rnn, bidirectional=False, reducer=reducer_lib.ConcatReducer(), **kwargs):
    """Initializes the layer.

    Args:
      rnn: The RNN layer to extend, built with ``return_sequences`` and
        ``return_state`` enabled.
      bidirectional: Make this layer bidirectional.
      reducer: A :class:`opennmt.layers.Reducer` instance to merge
        bidirectional states and outputs.
      **kwargs: Additional layer arguments.
    """
    super(_RNNWrapper, self).__init__(**kwargs)
    self.rnn = rnn
    self.reducer = reducer
    if bidirectional:
      self.rnn = tf.keras.layers.Bidirectional(self.rnn, merge_mode=None)

  def call(self, *args, **kwargs):  # pylint: disable=arguments-differ
    """Forwards the arguments to the RNN layer.

    Args:
      *args: Positional arguments of the RNN layer.
      **kwargs: Keyword arguments of the RNN layer.

    Returns:
      A tuple with the output sequences and the states.
    """
    outputs = self.rnn(*args, **kwargs)
    if isinstance(self.rnn, tf.keras.layers.Bidirectional):
      sequences = outputs[0:2]
      states = outputs[2:]
      fwd_states = states[:len(states)//2]
      bwd_states = states[len(states)//2:]
      if self.reducer is not None:
        sequences = self.reducer(sequences)
        states = tuple(self.reducer.zip_and_reduce(fwd_states, bwd_states))
      else:
        sequences = tuple(sequences)
        states = (fwd_states, bwd_states)
    else:
      sequences = outputs[0]
      states = tuple(outputs[1:])
    return sequences, states


class RNN(_RNNWrapper):
  """A simple RNN layer."""

  def __init__(self, cell, bidirectional=False, reducer=reducer_lib.ConcatReducer(), **kwargs):
    """Initializes the layer.

    Args:
      cell: The RNN cell to use.
      bidirectional: Make this layer bidirectional.
      reducer: A :class:`opennmt.layers.Reducer` instance to merge
        bidirectional states and outputs.
      **kwargs: Additional layer arguments.

    See Also:
      :func:`opennmt.layers.make_rnn_cell`
    """
    rnn = tf.keras.layers.RNN(cell, return_sequences=True, return_state=True)
    super(RNN, self).__init__(rnn, bidirectional=bidirectional, reducer=reducer, **kwargs)


class LSTM(tf.keras.layers.Layer):
  """A multi-layer LSTM.

  This differs from using :class:`opennmt.layers.RNN` with a ``LSTMCell`` in 2
  ways:

  - It uses ``tf.keras.layers.LSTM`` which is possibly accelerated by cuDNN on
    GPU.
  - Bidirectional outputs of each layer are reduced before feeding them to the
    next layer.
  """

  def __init__(self,
               num_layers,
               num_units,
               bidirectional=False,
               reducer=reducer_lib.ConcatReducer(),
               dropout=0,
               residual_connections=False,
               **kwargs):
    """Initializes the layer.

    Args:
      num_layers: Number of stacked LSTM layers.
      num_units: Dimension of the output space of each LSTM.
      bidirectional: Make each layer bidirectional.
      reducer: A :class:`opennmt.layers.Reducer` instance to merge
        the bidirectional states and outputs of each layer.
      dropout: The probability to drop units in each layer output.
      residual_connections: If ``True``, each layer input will be added to its
        output.
      **kwargs: Additional layer arguments.
    """
    super(LSTM, self).__init__(**kwargs)
    rnn_layers = [
        _RNNWrapper(
            tf.keras.layers.LSTM(num_units, return_sequences=True, return_state=True),
            bidirectional=bidirectional,
            reducer=reducer)
        for _ in range(num_layers)]
    self.layers = [
        common.LayerWrapper(
            layer,
            output_dropout=dropout,
            residual_connection=residual_connections)
        for layer in rnn_layers]

  def call(self, inputs, mask=None, training=None, initial_state=None):  # pylint: disable=arguments-differ
    all_states = []
    for i, layer in enumerate(self.layers):
      outputs, states = layer(
          inputs,
          mask=mask,
          training=training,
          initial_state=initial_state[i] if initial_state is not None else None)
      all_states.append(states)
      inputs = outputs
    return outputs, tuple(all_states)



"""Vocabulary utilities for Python scripts."""

import six

import tensorflow as tf
import numpy as np


class Vocab(object):
  """Vocabulary class."""

  def __init__(self, special_tokens=None):
    """Initializes a vocabulary.

    Args:
      special_tokens: A list of special tokens (e.g. start of sentence).
    """
    self._token_to_id = {}
    self._id_to_token = []
    self._frequency = []

    if special_tokens is not None:
      for index, token in enumerate(special_tokens):
        self._token_to_id[token] = index
        self._id_to_token.insert(index, token)

        # Set a very high frequency to avoid special tokens to be pruned. Note that Python sort
        # functions are stable which means that special tokens in pruned vocabularies will have
        # the same index.
        self._frequency.insert(index, float("inf"))

  @classmethod
  def from_file(cls, path, file_format="default"):
    """Creates from a vocabulary file.

    Args:
      path: The path to the vocabulary file.
      file_format: Define the format of the vocabulary file. Can be: default,
        sentencepiece. "default" is simply one token per line.

    Raises:
      ValueError: if :obj:`file_format` is invalid.
    """
    vocab = cls()
    vocab.load(path, file_format=file_format)
    return vocab

  @property
  def size(self):
    """Returns the number of entries of the vocabulary."""
    return len(self._id_to_token)

  @property
  def words(self):
    """Returns the list of words."""
    return self._id_to_token

  def __len__(self):
    """Returns the number of entries of the vocabulary."""
    return self.size

  def __contains__(self, token):
    """Returns ``True`` if the vocabulary contains :obj:`token`."""
    return self.lookup(token) is not None

  def add_from_text(self, filename, tokenizer=None):
    """Fills the vocabulary from a text file.

    Args:
      filename: The file to load from.
      tokenizer: A callable to tokenize a line of text.
    """
    with tf.io.gfile.GFile(filename, mode="rb") as text:
      for line in text:
        line = tf.compat.as_text(line.strip())
        if tokenizer:
          tokens = tokenizer.tokenize(line)
        else:
          tokens = line.split()
        for token in tokens:
          self.add(token)

  def serialize(self, path):
    """Writes the vocabulary on disk.

    Args:
      path: The path where the vocabulary will be saved.
    """
    with tf.io.gfile.GFile(path, mode="wb") as vocab:
      for token in self._id_to_token:
        vocab.write(tf.compat.as_bytes(token))
        vocab.write(b"\n")

  def load(self, path, file_format="default"):
    """Loads a serialized vocabulary.

    Args:
      path: The path to the vocabulary to load.
      file_format: Define the format of the vocabulary file. Can be: default,
        sentencepiece. "default" is simply one token per line.

    Raises:
      ValueError: if :obj:`file_format` is invalid.
    """
    with tf.io.gfile.GFile(path, mode="rb") as vocab:
      for line in vocab:
        token = line.rstrip(b"\r\n")
        if file_format == "default":
          self.add(token)
        elif file_format == "sentencepiece":
          token, _ = token.split(b"\t")
          if token in (b"<unk>", b"<s>", b"</s>"):  # Ignore SentencePiece special tokens.
            continue
          self.add(token)
        else:
          raise ValueError("Invalid vocabulary format: %s" % file_format)

  def add(self, token):
    """Adds a token or increases its frequency.

    Args:
      token: The string to add.
    """
    if isinstance(token, six.binary_type):
      token = tf.compat.as_text(token)
    if token not in self._token_to_id:
      index = self.size
      self._token_to_id[token] = index
      self._id_to_token.append(token)
      self._frequency.append(1)
    else:
      self._frequency[self._token_to_id[token]] += 1

  def lookup(self, identifier, default=None):
    """Lookups in the vocabulary.

    Args:
      identifier: A string or an index to lookup.
      default: The value to return if :obj:`identifier` is not found.

    Returns:
      The value associated with :obj:`identifier` or :obj:`default`.
    """
    value = None

    if isinstance(identifier, six.string_types):
      if isinstance(identifier, six.binary_type):
        identifier = tf.compat.as_text(identifier)
      if identifier in self._token_to_id:
        value = self._token_to_id[identifier]
    elif identifier < self.size:
      value = self._id_to_token[identifier]

    if value is None:
      return default
    else:
      return value

  def prune(self, max_size=0, min_frequency=1):
    """Creates a pruned version of the vocabulary.

    Args:
      max_size: The maximum vocabulary size.
      min_frequency: The minimum frequency of each entry.

    Returns:
      A new vocabulary.
    """
    sorted_ids = sorted(range(self.size), key=lambda k: self._frequency[k], reverse=True)
    new_size = len(sorted_ids)

    # Discard words that do not meet frequency requirements.
    for i in range(new_size - 1, 0, -1):
      index = sorted_ids[i]
      if self._frequency[index] < min_frequency:
        new_size -= 1
      else:
        break

    # Limit absolute size.
    if max_size > 0:
      new_size = min(new_size, max_size)

    new_vocab = Vocab()

    for i in range(new_size):
      index = sorted_ids[i]
      token = self._id_to_token[index]
      frequency = self._frequency[index]

      new_vocab._token_to_id[token] = i  # pylint: disable=protected-access
      new_vocab._id_to_token.append(token)  # pylint: disable=protected-access
      new_vocab._frequency.append(frequency)  # pylint: disable=protected-access

    return new_vocab

  def pad_to_multiple(self, multiple, num_oov_buckets=1):
    """Pads the vocabulary size to a multiple value.

    More specically, this method ensures that:

        ``(vocab_size + num_oov_buckets) % multiple == 0``

    Args:
      multiple: The multiple value.
      num_oov_buckets: The number of OOV buckets added during the training.
        Usually just 1 for the `<unk>` token.
    """
    i = 0
    while (self.size + num_oov_buckets) % multiple != 0:
      self.add("averyunlikelytoken%d" % i)
      i += 1


def get_mapping(current_vocab_path, new_vocab_path, mode="replace"):
  """Maps vocabulary new indices to old ones. -1 means that the entry is new."""
  mode = mode.lower()
  if mode not in ("merge", "replace"):
    raise ValueError("invalid vocab update mode: %s" % mode)
  current_vocab = Vocab.from_file(current_vocab_path)
  new_vocab = Vocab.from_file(new_vocab_path)
  mapping = []
  if mode == "merge":
    final_vocab = Vocab.from_file(current_vocab_path)
    mapping = [i for i in range(current_vocab.size)]
    for new_word in new_vocab.words:
      if current_vocab.lookup(new_word) is None:
        mapping.append(-1)
        final_vocab.add(new_word)
  elif mode == "replace":
    final_vocab = new_vocab
    for new_word in new_vocab.words:
      idx = current_vocab.lookup(new_word)
      if idx is not None:
        mapping.append(idx)
      else:
        mapping.append(-1)
  mapping.append(current_vocab.size)  # <unk> token is always the last entry.
  return mapping, final_vocab

def update_variable(ref_variable, new_variable, mapping, vocab_axis=0):
  """Update a vocabulary variable, possibly copying previous entries based on
  mapping.
  """
  ref = ref_variable.numpy()
  new = new_variable.numpy()
  perm = None
  if vocab_axis != 0:
    # Make the dimension to index the first.
    perm = list(range(len(ref.shape)))
    perm[0], perm[vocab_axis] = perm[vocab_axis], perm[0]
    ref = np.transpose(ref, axes=perm)
    new = np.transpose(new, axes=perm)
  for i, j in enumerate(mapping):
    if j >= 0:
      new[i] = ref[j]
  if perm is not None:
    new = np.transpose(new, axes=perm)
  new_variable.assign(new)
  return new_variable

def update_variable_and_slots(ref_variable,
                              new_variable,
                              ref_optimizer,
                              new_optimizer,
                              mapping,
                              vocab_axis=0):
  """Update a vocabulary variable and its associated optimizer slots (if any)."""
  variables = []
  variables.append(update_variable(ref_variable, new_variable, mapping, vocab_axis=vocab_axis))
  ref_slot_names = ref_optimizer.get_slot_names()
  new_slot_names = new_optimizer.get_slot_names()
  for slot_name in ref_slot_names:
    if slot_name not in new_slot_names:
      continue
    ref_slot = ref_optimizer.get_slot(ref_variable, slot_name)
    new_slot = new_optimizer.get_slot(new_variable, slot_name)
    variables.append(update_variable(ref_slot, new_slot, mapping, vocab_axis=vocab_axis))
  return variables



"""Module defining various utilities."""

from opennmt.utils.checkpoint import average_checkpoints

from opennmt.utils.decoding import BeamSearch
from opennmt.utils.decoding import BestSampler
from opennmt.utils.decoding import DecodingResult
from opennmt.utils.decoding import DecodingStrategy
from opennmt.utils.decoding import GreedySearch
from opennmt.utils.decoding import RandomSampler
from opennmt.utils.decoding import Sampler
from opennmt.utils.decoding import dynamic_decode

from opennmt.utils.losses import cross_entropy_loss
from opennmt.utils.losses import cross_entropy_sequence_loss
from opennmt.utils.losses import guided_alignment_cost
from opennmt.utils.losses import max_margin_loss
from opennmt.utils.losses import regularization_penalty

from opennmt.utils.misc import format_translation_output

from opennmt.utils.scorers import BLEUScorer
from opennmt.utils.scorers import ROUGEScorer
from opennmt.utils.scorers import Scorer

from opennmt.utils.tensor import roll_sequence



"""Module defining decoders."""

from opennmt.decoders.decoder import Decoder
from opennmt.decoders.decoder import get_sampling_probability

from opennmt.decoders.rnn_decoder import AttentionalRNNDecoder
from opennmt.decoders.rnn_decoder import RNMTPlusDecoder
from opennmt.decoders.rnn_decoder import RNNDecoder

from opennmt.decoders.self_attention_decoder import SelfAttentionDecoder



"""Standalone script to generate word vocabularies from monolingual corpus."""

import argparse

from opennmt import constants
from opennmt import tokenizers
from opennmt import data


def main():
  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument(
      "data", nargs="*",
      help="Source text file.")
  parser.add_argument(
      "--from_vocab", default=None,
      help="Build from a saved vocabulary (see also --from_format).")
  parser.add_argument(
      "--from_format", default="default", choices=["default", "sentencepiece"],
      help="The format of the saved vocabulary (see also --from_vocab).")
  parser.add_argument(
      "--save_vocab", required=True,
      help="Output vocabulary file.")
  parser.add_argument(
      "--min_frequency", type=int, default=1,
      help="Minimum word frequency.")
  parser.add_argument(
      "--size", type=int, default=0,
      help="Maximum vocabulary size. If = 0, do not limit vocabulary.")
  parser.add_argument(
      "--size_multiple", type=int, default=1,
      help=("Ensure that the vocabulary size + 1 is a multiple of this value "
            "(+ 1 represents the <unk> token that will be added during the training."))
  parser.add_argument(
      "--without_sequence_tokens", default=False, action="store_true",
      help="If set, do not add special sequence tokens (start, end) in the vocabulary.")
  parser.add_argument(
      "--tokenizer_config", default=None,
      help="Tokenization configuration.")
  parser.add_argument(
      "--sentencepiece", nargs="*", default=None,
      help=("Build a SentencePiece model and vocabulary. This option accepts additional "
            "training parameters (e.g. --sentencepiece character_coverage=0.98)."))
  args = parser.parse_args()

  special_tokens = [constants.PADDING_TOKEN]
  if not args.without_sequence_tokens:
    special_tokens.append(constants.START_OF_SENTENCE_TOKEN)
    special_tokens.append(constants.END_OF_SENTENCE_TOKEN)

  vocab = data.Vocab(special_tokens=special_tokens)
  num_oov_buckets = 1

  if args.sentencepiece is not None:
    import pyonmttok
    if args.size_multiple == 1:
      vocab_size = args.size
    else:
      # Round vocabulary size to the next multiple of args.size_multiple
      vocab_size = (
          args.size - (args.size + num_oov_buckets) % args.size_multiple + args.size_multiple)
    sp_params = dict(map(lambda arg: tuple(arg.split("=")), args.sentencepiece))
    sp_trainer = pyonmttok.SentencePieceLearner(
        keep_vocab=True, vocab_size=vocab_size, **sp_params)
    for data_file in args.data:
      sp_trainer.ingest_file(data_file)
    sp_trainer.learn(args.save_vocab, verbose=True)
    args.save_vocab = args.save_vocab + ".vocab"
    vocab.load(args.save_vocab, file_format="sentencepiece")
  else:
    if args.from_vocab is not None:
      vocab.load(args.from_vocab, file_format=args.from_format)
    tokenizer = tokenizers.make_tokenizer(args.tokenizer_config)
    for data_file in args.data:
      vocab.add_from_text(data_file, tokenizer=tokenizer)
    vocab = vocab.prune(max_size=args.size, min_frequency=args.min_frequency)
    vocab.pad_to_multiple(args.size_multiple, num_oov_buckets=num_oov_buckets)

  vocab.serialize(args.save_vocab)


if __name__ == "__main__":
  main()



"""Define learning rate decay functions."""

import sys

import tensorflow as tf
import numpy as np


def make_learning_rate_schedule(initial_learning_rate,
                                schedule_type,
                                schedule_params=None,
                                schedule_step_duration=1,
                                start_step=0,
                                minimum_learning_rate=0):
  """Creates the learning rate schedule.

  Args:
    initial_learning_rate: The initial learning rate value or scale.
    schedule_type: The type of decay. A function from
      ``tf.keras.optimizers.schedules``
      or :mod:`opennmt.schedules` as a string.
    schedule_params: Additional parameters for the decay function.
    schedule_step_duration: The number of training steps that make 1 decay step.
    start_step: Start the schedule after this many steps.
    minimum_learning_rate: Do not decay past this learning rate value.

  Returns:
    A ``tf.keras.optimizers.schedules.LearningRateSchedule`` instance.

  Raises:
    ValueError: if :obj:`schedule_type` can not be resolved to an existing
      schedule.

  See Also:
    :class:`opennmt.schedules.ScheduleWrapper`
  """
  schedule_name = None
  if schedule_name is None:
    schedule_name = getattr(tf.keras.optimizers.schedules, schedule_type, None)
  if schedule_name is None:
    schedule_name = getattr(sys.modules[__name__], schedule_type, None)
  if schedule_name is None:
    raise ValueError("Unknown learning rate schedule: {}".format(schedule_type))

  if schedule_params is None:
    schedule_params = {}
  schedule = schedule_name(initial_learning_rate, **schedule_params)
  schedule = ScheduleWrapper(
      schedule,
      step_start=start_step,
      step_duration=schedule_step_duration,
      minimum_learning_rate=minimum_learning_rate)
  return schedule


class ScheduleWrapper(tf.keras.optimizers.schedules.LearningRateSchedule):
  """Wrapper to augment a learning rate scheduler behavior."""

  def __init__(self,
               schedule,
               step_start=0,
               step_duration=1,
               minimum_learning_rate=0):
    """Initializes the decay function.

    Args:
      schedule: A ``tf.keras.optimizers.schedules.LearningRateSchedule``.
      step_duration: The number of training steps that make 1 decay step.
      start_step: Start decay after this many steps.
      minimum_learning_rate: Do not decay past this learning rate value.

    See Also:
      :class:`opennmt.schedules.make_learning_rate_schedule`
    """
    self.schedule = schedule
    self.step_start = step_start
    self.step_duration = step_duration
    self.minimum_learning_rate = minimum_learning_rate

  def __call__(self, step):
    # Map the training step to a decay step.
    step = tf.maximum(step - self.step_start, 0)
    step //= self.step_duration
    learning_rate = self.schedule(step)
    return tf.maximum(learning_rate, self.minimum_learning_rate)


class NoamDecay(tf.keras.optimizers.schedules.LearningRateSchedule):
  """Defines the decay function described in https://arxiv.org/abs/1706.03762."""

  def __init__(self, scale, model_dim, warmup_steps):
    """Initializes the decay function.

    Args:
      scale: The scale constant.
      model_dim: The model dimension.
      warmup_steps: The number of warmup steps.
    """
    self.scale = tf.cast(scale, tf.float32)
    self.model_dim = tf.cast(model_dim, tf.float32)
    self.warmup_steps = tf.cast(warmup_steps, tf.float32)

  def __call__(self, step):
    step = tf.cast(step + 1, tf.float32)
    return (self.scale
            * tf.pow(self.model_dim, -0.5)
            * tf.minimum(tf.pow(step, -0.5), step * tf.pow(self.warmup_steps, -1.5)))


class RsqrtDecay(tf.keras.optimizers.schedules.LearningRateSchedule):
  """Decay based on the reciprocal of the step square root."""

  def __init__(self, scale, warmup_steps):
    """Initializes the decay function.

    Args:
      scale: The scale constant.
      warmup_steps: The number of warmup steps.
    """
    self.scale = tf.cast(scale, tf.float32)
    self.warmup_steps = tf.cast(warmup_steps, tf.float32)

  def __call__(self, step):
    step = tf.cast(step, tf.float32)
    return self.scale * tf.math.rsqrt(tf.maximum(step, self.warmup_steps))


class CosineAnnealing(tf.keras.optimizers.schedules.LearningRateSchedule):
  """Decay using a cosine annealing schedule."""

  def __init__(self, eta_max, eta_min=0, max_step=1000000, warmup_steps=None):
    """Initializes the decay function.

    Args:
      eta_max: Maximum learning rate.
      eta_min: Minimum learning rate.
      max_step: The last step of the scedule.
      warmup_steps: The number of steps to increment the learning rate linearly
        from 0 to :obj:`scale` before annealing.
    """
    self.eta_max = tf.cast(eta_max, tf.float32)
    self.eta_min = tf.cast(eta_min, tf.float32)
    self.max_step = tf.cast(max_step, tf.float32)
    self.warmup_steps = tf.cast(warmup_steps, tf.float32) if warmup_steps is not None else None

  def __call__(self, step):
    step = tf.cast(step, tf.float32)
    annealing = lambda: (
        self.eta_min
        + 0.5 * (self.eta_max - self.eta_min) * (1 + tf.cos(np.pi * step / self.max_step)))
    linear = lambda: self.eta_max * step / tf.cast(self.warmup_steps, tf.float32)
    if self.warmup_steps is None:
      return annealing()
    return tf.cond(tf.less(step, self.warmup_steps), true_fn=linear, false_fn=annealing)


class RNMTPlusDecay(tf.keras.optimizers.schedules.LearningRateSchedule):
  """Defines the decay function described in https://arxiv.org/abs/1804.09849."""

  def __init__(self,
               scale,
               num_replicas,
               warmup_steps=500,
               start_step=600000,
               end_step=1200000):
    """Initializes the decay function.

    Args:
      scale: The scale constant.
      num_replicas: The number of concurrent model replicas.
      warmup_steps: The number of warmup steps.
      start_step: The start step of the exponential decay.
      end_step: The end step of the exponential decay.
    """
    self.scale = tf.cast(scale, tf.float32)
    self.num_replicas = tf.cast(num_replicas, tf.float32)
    self.warmup_steps = tf.cast(warmup_steps, tf.float32)
    self.start_step = tf.cast(start_step, tf.float32)
    self.end_step = tf.cast(end_step, tf.float32)

  def __call__(self, step):
    t = tf.cast(step, tf.float32)
    n = self.num_replicas
    p = self.warmup_steps
    s = self.start_step
    e = self.end_step
    return self.scale * tf.minimum(
        tf.minimum(1 + (t * (n - 1)) / (n * p), n),
        n * tf.pow(2 * n, (s - n * t) / (e - s)))



"""ARK data file to TFRecords converter.

The scripts takes the ARK data file and optionally the indexed target text
to write aligned source and target data.
"""

from __future__ import print_function

import argparse
import io
import numpy as np
import tensorflow as tf

from opennmt.inputters.record_inputter import write_sequence_record


def consume_next_vector(ark_file):
  """Consumes the next vector.

  Args:
    ark_file: The ARK data file.

  Returns:
    The next vector as a 2D Numpy array.
  """
  idx = None
  vector = []

  for line in ark_file:
    line = line.strip()
    fields = line.split()

    if not idx:
      idx = fields[0]
      fields.pop(0)
      fields.pop(0)

    end = fields and fields[-1] == "]"

    if end:
      fields.pop()

    if fields:
      vector.append(fields)

    if end:
      break

  return idx, np.asarray(vector, dtype=tf.float32)

def consume_next_text(text_file):
  """Consumes the next text line from `text_file`."""
  idx = None
  text = text_file.readline()

  if text:
    tokens = text.strip().split()
    idx = tokens[0]
    tokens.pop(0)
    text = " ".join(tokens)

  return idx, text

def write_text(text, writer):
  """Serializes a line of text."""
  writer.write(text)
  writer.write("\n")

def ark_to_records_aligned(ark_filename, text_filename, out_prefix, compression_type=None):
  """Converts ARK and text datasets to aligned TFRecords and text datasets."""
  record_filename = "%s.records" % out_prefix
  if compression_type == "GZIP":
    record_filename = "%s.gz" % record_filename
  record_writer = tf.io.TFRecordWriter(record_filename, options=compression_type)
  text_writer = io.open(out_prefix + ".txt", encoding="utf-8", mode="w")

  ark_buffer = {}
  text_buffer = {}
  count = 0

  def _write_example(vector, text):
    write_sequence_record(vector, record_writer)
    write_text(text, text_writer)

  def _search_aligned():
    for idx in ark_buffer:
      if idx in text_buffer:
        vector = ark_buffer[idx]
        text = text_buffer[idx]

        del ark_buffer[idx]
        del text_buffer[idx]

        return vector, text

    return None, None

  with io.open(ark_filename, encoding="utf-8") as ark_file, open(text_filename, encoding="utf-8") as text_file: #pylint: disable=line-too-long
    while True:
      ark_idx, vector = consume_next_vector(ark_file)
      text_idx, text = consume_next_text(text_file)

      if not ark_idx and not text_idx:
        # Both files are empty.
        break

      if ark_idx == text_idx:
        # If the indices match, write the example.
        _write_example(vector, text)
        count += 1
      else:
        # Otherwise store the entries.
        if ark_idx:
          ark_buffer[ark_idx] = vector
        if text_idx:
          text_buffer[text_idx] = text

        # Look if we can now find aligned entries.
        vector, text = _search_aligned()

        if vector is not None:
          _write_example(vector, text)
          count += 1

  # Search alignments in stored entries.
  while True:
    vector, text = _search_aligned()
    if vector is None:
      break
    _write_example(vector, text)
    count += 1

  record_writer.close()
  text_writer.close()

  print("Saved {} aligned records.".format(count))

def ark_to_records(ark_filename, out_prefix, compression_type=None):
  """Converts ARK dataset to TFRecords."""
  record_writer = tf.io.TFRecordWriter(out_prefix + ".records", options=compression_type)
  count = 0

  with io.open(ark_filename, encoding="utf-8") as ark_file:
    while True:
      ark_idx, vector = consume_next_vector(ark_file)
      if not ark_idx:
        break
      write_sequence_record(vector, record_writer)
      count += 1

  record_writer.close()
  print("Saved {} records.".format(count))


def main():
  parser = argparse.ArgumentParser()
  parser.add_argument("--ark", required=True,
                      help="Indexed ARK data file.")
  parser.add_argument("--txt",
                      help=("Indexed target text data file "
                            "(must set it to align source and target files)."))
  parser.add_argument("--out", required=True,
                      help="Output files prefix (will be suffixed by .records and .txt).")
  parser.add_argument("--compression_type", default=None, choices=["GZIP"],
                      help="Optional compression type.")
  args = parser.parse_args()

  if args.txt:
    ark_to_records_aligned(args.ark, args.txt, args.out, compression_type=args.compression_type)
  else:
    ark_to_records(args.ark, args.out, compression_type=args.compression_type)

if __name__ == "__main__":
  main()



"""Dynamic decoding utilities."""

import abc
import collections
import six

import tensorflow as tf
import tensorflow_addons as tfa

from opennmt import constants
from opennmt.utils import misc


@six.add_metaclass(abc.ABCMeta)
class Sampler(object):
  """Base class for samplers."""

  @abc.abstractmethod
  def __call__(self, scores, num_samples=1):
    """Samples predictions.

    Args:
      scores: The scores to sample from, a tensor of shape
        ``[batch_size, vocab_size]``.
      num_samples: The number of samples per batch to produce.

    Returns:
      A tuple ``(sample_ids, sample_scores)``.
    """
    raise NotImplementedError()

  @staticmethod
  def from_params(params):
    """Constructs a sampler based on user parameters.

    Args:
      params: A dictionary of user parameters.

    Returns:
      A :class:`opennmt.utils.Sampler` instance.
    """
    sampling_topk = params.get("sampling_topk", 1)
    if sampling_topk == 1:
      return BestSampler()
    else:
      return RandomSampler(
          from_top_k=sampling_topk,
          temperature=params.get("sampling_temperature"))

class RandomSampler(Sampler):
  """Randomly samples from model outputs."""

  def __init__(self, from_top_k=None, temperature=None):
    """Initializes the random sampler.

    Args:
      from_top_k: Sample from the top K predictions instead of the full
        distribution.
      temperature: Divide logits by this value. High temperatures generate more
        random samples.
    """
    if from_top_k is not None and from_top_k <= 0:
      from_top_k = None
    self.from_top_k = from_top_k
    self.temperature = temperature

  def __call__(self, scores, num_samples=1):
    if self.from_top_k is None:
      sample_ids = _sample_from(scores, num_samples, temperature=self.temperature)
    else:
      top_scores, top_ids = tf.nn.top_k(scores, k=self.from_top_k)
      sample_ids = _sample_from(top_scores, num_samples, temperature=self.temperature)
      sample_ids = _gather_from_word_indices(top_ids, sample_ids)
    sample_scores = _gather_from_word_indices(scores, sample_ids)
    return sample_ids, sample_scores

class BestSampler(Sampler):
  """Sample the best predictions."""

  def __call__(self, scores, num_samples=1):
    sample_scores, sample_ids = tf.nn.top_k(scores, k=num_samples)
    return sample_ids, sample_scores


@six.add_metaclass(abc.ABCMeta)
class DecodingStrategy(object):
  """Base class for decoding strategies."""

  @property
  def num_hypotheses(self):
    """The number of hypotheses returned by this strategy."""
    return 1

  @staticmethod
  def from_params(params):
    """Constructs a decoding strategy based on user parameters.

    Args:
      params: A dictionary of user parameters.

    Returns:
      A :class:`opennmt.utils.DecodingStrategy` instance.
    """
    beam_size = params.get("beam_width", 1)
    if beam_size > 1:
      return BeamSearch(
          beam_size,
          length_penalty=params.get("length_penalty", 0),
          coverage_penalty=params.get("coverage_penalty", 0))
    else:
      return GreedySearch()

  @abc.abstractmethod
  def _initialize(self, batch_size, start_ids, attention_size=None):
    """Initializes the strategy.

    Args:
      batch_size: The batch size.
      start_ids: The start decoding ids.
      attention_size: If known, the size of the attention vectors (i.e. the
        maximum source length).

    Returns:
      A tuple containing,

      - The (possibly transformed) start decoding ids.
      - The tensor of finished flags.
      - Initial log probabilities per batch.
      - A sequence of additional tensors used during the decoding.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def _step(self,
            step,
            sampler,
            log_probs,
            cum_log_probs,
            finished,
            state,
            extra_vars,
            attention=None):
    """Updates the strategy state.

    Args:
      step: The current decoding step.
      sampler: The sampler that produces predictions.
      log_probs: The model log probabilities.
      cum_log_probs: The cumulated log probabilities per batch.
      finished: The current finished flags.
      state: The decoder state.
      extra_vars: Additional tensors from this decoding strategy.
      attention: The attention vector for the current step.

    Returns:
      A tuple containing,

      - The predicted word ids.
      - The new cumulated log probabilities.
      - The updated finished flags.
      - The update decoder state.
      - Additional tensors from this decoding strategy.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def _finalize(self, outputs, end_id, extra_vars, attention=None):
    """Finalize the predictions.

    Args:
      outputs: The array of sampled ids.
      end_id: The end token id.
      extra_vars: Additional tensors from this decoding strategy.
      attention: The array of attention outputs.

    Returns:
      A tuple containing,

      - The final predictions as a tensor of shape [B, H, T].
      - The final attention history of shape [B, H, T, S].
      - The final sequence lengths of shape [B, H].
    """
    raise NotImplementedError()


class GreedySearch(DecodingStrategy):
  """A basic greedy search strategy."""

  def _initialize(self, batch_size, start_ids, attention_size=None):
    finished = tf.zeros([batch_size], dtype=tf.bool)
    initial_log_probs = tf.zeros([batch_size], dtype=tf.float32)
    return start_ids, finished, initial_log_probs, []

  def _step(self,
            step,
            sampler,
            log_probs,
            cum_log_probs,
            finished,
            state,
            extra_vars,
            attention=None):
    sample_ids, sample_log_probs = sampler(log_probs)
    sample_ids = tf.reshape(sample_ids, [-1])
    sample_log_probs = tf.reshape(sample_log_probs, [-1])
    cum_log_probs += sample_log_probs
    return sample_ids, cum_log_probs, finished, state, extra_vars

  def _finalize(self, outputs, end_id, extra_vars, attention=None):
    ids = tf.transpose(outputs.stack())
    ids = tf.expand_dims(ids, 1)
    lengths = _lengths_from_ids(ids, end_id)
    if attention is not None:
      attention = tf.transpose(attention.stack(), perm=[1, 0, 2])
      attention = tf.expand_dims(attention, 1)
    return ids, attention, lengths


class BeamSearch(DecodingStrategy):
  """A beam search strategy."""

  def __init__(self, beam_size, length_penalty=0, coverage_penalty=0):
    """Initializes the decoding strategy.

    Args:
      beam_size: The number of paths to consider per batch.
      length_penalty: Length penalty, see https://arxiv.org/abs/1609.08144.
      coverage_penalty: Coverage penalty, see https://arxiv.org/abs/1609.08144.
    """
    self.beam_size = beam_size
    self.length_penalty = length_penalty
    self.coverage_penalty = coverage_penalty

  @property
  def num_hypotheses(self):
    return self.beam_size

  def _initialize(self, batch_size, start_ids, attention_size=None):
    start_ids = tfa.seq2seq.tile_batch(start_ids, self.beam_size)
    finished = tf.zeros([batch_size * self.beam_size], dtype=tf.bool)
    # Give all probability to first beam for the first iteration.
    initial_log_probs = tf.tile([0.] + [-float("inf")] * (self.beam_size - 1), [batch_size])
    parent_ids = tf.TensorArray(tf.int32, size=0, dynamic_size=True)
    sequence_lengths = tf.zeros([batch_size * self.beam_size], dtype=tf.int32)
    extra_vars = [parent_ids, sequence_lengths]
    if self.coverage_penalty != 0:
      if attention_size is None:
        raise ValueError("The attention size should be known to support coverage penalty")
      accumulated_attention = tf.zeros([batch_size * self.beam_size, attention_size])
      extra_vars.append(accumulated_attention)
    return start_ids, finished, initial_log_probs, tuple(extra_vars)

  def _get_scores(self, log_probs, sequence_lengths, finished, accumulated_attention=None):
    scores = log_probs
    if self.length_penalty != 0:
      expand_sequence_lengths = tf.expand_dims(sequence_lengths, 1)
      scores /= tf.pow(((5. + tf.cast(expand_sequence_lengths + 1, scores.dtype)) / 6.),
                       self.length_penalty)
    if self.coverage_penalty != 0:
      # Mask out of range steps with ones (log(1) == 0).
      accumulated_attention = tf.where(
          tf.equal(accumulated_attention, 0.0),
          x=tf.ones_like(accumulated_attention),
          y=accumulated_attention)
      coverage_penalty = tf.reduce_sum(
          tf.math.log(tf.minimum(accumulated_attention, 1.0)), 1)
      # Apply coverage penalty to finished predictions.
      coverage_penalty *= tf.cast(finished, coverage_penalty.dtype)
      scores += self.coverage_penalty * tf.expand_dims(coverage_penalty, 1)
    return scores

  def _step(self,
            step,
            sampler,
            log_probs,
            cum_log_probs,
            finished,
            state,
            extra_vars,
            attention=None):
    parent_ids = extra_vars[0]
    sequence_lengths = extra_vars[1]

    if self.coverage_penalty != 0:
      if attention is None:
        raise ValueError("Coverage penalty is enabled but the model did not "
                         "return an attention vector")
      not_finished = tf.math.logical_not(finished)
      attention *= tf.expand_dims(tf.cast(not_finished, attention.dtype), 1)
      accumulated_attention = extra_vars[2] + attention
    else:
      accumulated_attention = None

    # Compute scores from log probabilities.
    vocab_size = log_probs.shape[-1]
    total_probs = log_probs + tf.expand_dims(cum_log_probs, 1)  # Add current beam probability.
    scores = self._get_scores(
        total_probs,
        sequence_lengths,
        finished,
        accumulated_attention=accumulated_attention)
    scores = tf.reshape(scores, [-1, self.beam_size * vocab_size])
    total_probs = tf.reshape(total_probs, [-1, self.beam_size * vocab_size])

    # Sample predictions.
    sample_ids, sample_scores = sampler(scores, num_samples=self.beam_size)
    sample_ids = tf.reshape(sample_ids, [-1])
    sample_scores = tf.reshape(sample_scores, [-1])

    # Resolve beam origin and word ids.
    word_ids = sample_ids % vocab_size
    beam_ids = sample_ids // vocab_size
    beam_indices = (
        (tf.range(tf.shape(word_ids)[0]) // self.beam_size)
        * self.beam_size + beam_ids)

    # Update sequence_length of unfinished sequence.
    sequence_lengths = tf.where(finished, x=sequence_lengths, y=sequence_lengths + 1)

    # Update state and flags.
    cum_log_probs = _gather_from_word_indices(total_probs, sample_ids)
    finished = tf.gather(finished, beam_indices)
    sequence_lengths = tf.gather(sequence_lengths, beam_indices)
    parent_ids = parent_ids.write(step, beam_ids)
    extra_vars = [parent_ids, sequence_lengths]
    if accumulated_attention is not None:
      accumulated_attention = tf.gather(accumulated_attention, beam_indices)
      extra_vars.append(accumulated_attention)
    state = tf.nest.map_structure(lambda s: _gather_state(s, beam_indices), state)
    return word_ids, cum_log_probs, finished, state, tuple(extra_vars)

  def _finalize(self, outputs, end_id, extra_vars, attention=None):
    parent_ids = extra_vars[0]
    sequence_lengths = extra_vars[1]
    maximum_lengths = tf.reduce_max(tf.reshape(sequence_lengths, [-1, self.beam_size]), axis=-1)
    max_time = outputs.size()
    array_shape = [max_time, -1, self.beam_size]
    step_ids = tf.reshape(outputs.stack(), array_shape)
    parent_ids = tf.reshape(parent_ids.stack(), array_shape)
    ids = tfa.seq2seq.gather_tree(step_ids, parent_ids, maximum_lengths, end_id)
    ids = tf.transpose(ids, perm=[1, 2, 0])
    lengths = _lengths_from_ids(ids, end_id)
    if attention is not None:
      attention = tfa.seq2seq.gather_tree_from_array(
          attention.stack(), parent_ids, lengths)
      attention = tf.transpose(attention, perm=[1, 0, 2])
      attention = tf.reshape(
          attention, [tf.shape(ids)[0], self.beam_size, max_time, -1])
    return ids, attention, lengths


class DecodingResult(
    collections.namedtuple("DecodingResult",
                           ("ids", "lengths", "log_probs", "attention", "state"))):
  """Final decoding result.

  Args:
    ids: The predicted ids of shape :math:`[B, H, T]`.
    lengths: The produced sequences length of shape :math:`[B, H]`.
    log_probs: The cumulated log probabilities of shape :math:`[B, H]`.
    attention: The attention history of shape :math:`[B, H, T_t, T_s]`.
    state: The final decoding state.
  """


def dynamic_decode(symbols_to_logits_fn,
                   start_ids,
                   end_id=constants.END_OF_SENTENCE_ID,
                   initial_state=None,
                   decoding_strategy=None,
                   sampler=None,
                   maximum_iterations=None,
                   minimum_iterations=0,
                   attention_history=False,
                   attention_size=None):
  """Dynamic decoding.

  Args:
    symbols_to_logits_fn: A callable taking ``(symbols, step, state)`` and
      returning ``(logits, state, attention)`` (``attention`` is optional).
    start_ids: Initial input IDs of shape :math:`[B]`.
    end_id: ID of the end of sequence token.
    initial_state: Initial decoder state.
    decoding_strategy: A :class:`opennmt.utils.DecodingStrategy`
      instance that define the decoding logic. Defaults to a greedy search.
    sampler: A :class:`opennmt.utils.Sampler` instance that samples
      predictions from the model output. Defaults to an argmax sampling.
    maximum_iterations: The maximum number of iterations to decode for.
    minimum_iterations: The minimum number of iterations to decode for.
    attention_history: Gather attention history during the decoding.
    attention_size: If known, the size of the attention vectors (i.e. the
      maximum source length).

  Returns:
    A :class:`opennmt.utils.DecodingResult` instance.
  """
  if initial_state is None:
    initial_state = {}
  if decoding_strategy is None:
    decoding_strategy = GreedySearch()
  if sampler is None:
    sampler = BestSampler()

  def _cond(step, finished, state, inputs, outputs, attention, cum_log_probs, extra_vars):  # pylint: disable=unused-argument
    return tf.reduce_any(tf.logical_not(finished))

  def _body(step, finished, state, inputs, outputs, attention, cum_log_probs, extra_vars):
    # Get log probs from the model.
    result = symbols_to_logits_fn(inputs, step, state)
    logits, state = result[0], result[1]
    attn = result[2] if len(result) > 2 else None
    logits = tf.cast(logits, tf.float32)

    # Penalize or force EOS.
    batch_size, vocab_size = misc.shape_list(logits)
    eos_max_prob = tf.one_hot(
        tf.fill([batch_size], end_id),
        vocab_size,
        on_value=logits.dtype.max,
        off_value=logits.dtype.min)
    logits = tf.cond(
        step < minimum_iterations,
        true_fn=lambda: _penalize_token(logits, end_id),
        false_fn=lambda: tf.where(
            tf.broadcast_to(tf.expand_dims(finished, -1), tf.shape(logits)),
            x=eos_max_prob,
            y=logits))
    log_probs = tf.nn.log_softmax(logits)

    # Run one decoding strategy step.
    output, next_cum_log_probs, finished, state, extra_vars = (
        decoding_strategy._step(  # pylint: disable=protected-access
            step,
            sampler,
            log_probs,
            cum_log_probs,
            finished,
            state,
            extra_vars,
            attention=attn))

    # Update loop vars.
    if attention_history:
      if attn is None:
        raise ValueError("attention_history is set but the model did not return attention")
      attention = attention.write(step, tf.cast(attn, tf.float32))
    outputs = outputs.write(step, output)
    cum_log_probs = tf.where(finished, x=cum_log_probs, y=next_cum_log_probs)
    finished = tf.logical_or(finished, tf.equal(output, end_id))
    return step + 1, finished, state, output, outputs, attention, cum_log_probs, extra_vars

  start_ids = tf.convert_to_tensor(start_ids)
  batch_size = tf.shape(start_ids)[0]
  ids_dtype = start_ids.dtype
  start_ids = tf.cast(start_ids, tf.int32)
  start_ids, finished, initial_log_probs, extra_vars = (
      decoding_strategy._initialize(  # pylint: disable=protected-access
          batch_size, start_ids, attention_size=attention_size))
  step = tf.constant(0, dtype=tf.int32)
  outputs = tf.TensorArray(tf.int32, size=0, dynamic_size=True)
  attention = tf.TensorArray(tf.float32, size=0, dynamic_size=True)

  _, _, state, _, outputs, attention, log_probs, extra_vars = tf.while_loop(
      _cond,
      _body,
      loop_vars=(
          step,
          finished,
          initial_state,
          start_ids,
          outputs,
          attention,
          initial_log_probs,
          extra_vars),
      shape_invariants=(
          step.shape,
          finished.shape,
          tf.nest.map_structure(_get_shape_invariants, initial_state),
          start_ids.shape,
          tf.TensorShape(None),
          tf.TensorShape(None),
          initial_log_probs.shape,
          tf.nest.map_structure(_get_shape_invariants, extra_vars)),
      parallel_iterations=1,
      back_prop=False,
      maximum_iterations=maximum_iterations)

  ids, attention, lengths = decoding_strategy._finalize(  # pylint: disable=protected-access
      outputs,
      end_id,
      extra_vars,
      attention=attention if attention_history else None)
  if attention is not None:
    attention = attention[:, :, :-1]  # Ignore attention for </s>.
  log_probs = tf.reshape(log_probs, [batch_size, decoding_strategy.num_hypotheses])
  ids = tf.cast(ids, ids_dtype)
  return DecodingResult(
      ids=ids,
      lengths=lengths,
      log_probs=log_probs,
      attention=attention,
      state=state)


def _gather_state(tensor, indices):
  """Gather batch indices from the tensor."""
  if isinstance(tensor, tf.TensorArray) or tensor.shape.ndims == 0:
    return tensor
  return tf.gather(tensor, indices)

def _get_shape_invariants(tensor):
  """Returns the shape of the tensor but sets middle dims to None."""
  if isinstance(tensor, tf.TensorArray):
    shape = None
  else:
    shape = tensor.shape.as_list()
    for i in range(1, len(shape) - 1):
      shape[i] = None
  return tf.TensorShape(shape)

def _penalize_token(log_probs, token_id, penalty=-1e7):
  """Penalize token probabilities."""
  depth = log_probs.shape[-1]
  penalty = tf.one_hot([token_id], depth, on_value=tf.cast(penalty, log_probs.dtype))
  return log_probs + penalty

def _sample_from(logits, num_samples, temperature=None):
  """Sample N values from the unscaled probability distribution."""
  if temperature is not None:
    logits /= tf.cast(temperature, logits.dtype)
  return tf.random.categorical(logits, num_samples, dtype=tf.int32)

def _gather_from_word_indices(tensor, indices):
  """Index the depth dim of a 2D tensor."""
  output_shape = misc.shape_list(indices)
  batch_size = tf.shape(tensor)[0]
  num_indices = tf.size(indices) // batch_size
  batch_pos = tf.range(batch_size * num_indices) // num_indices
  tensor = tf.gather_nd(tensor, tf.stack([batch_pos, tf.reshape(indices, [-1])], axis=-1))
  tensor = tf.reshape(tensor, output_shape)
  return tensor

def _lengths_from_ids(ids, end_id):
  """Compute sequence lengths from word ids."""
  lengths = tf.not_equal(ids, end_id)
  lengths = tf.cast(lengths, tf.int32)
  lengths = tf.reduce_sum(lengths, axis=-1)
  return lengths



"""Define the self-attention encoder."""

import tensorflow as tf

from opennmt.layers import transformer

from opennmt.encoders.encoder import Encoder
from opennmt.layers.position import SinusoidalPositionEncoder
from opennmt.layers import common


class SelfAttentionEncoder(Encoder):
  """Encoder using self-attention as described in
  https://arxiv.org/abs/1706.03762.
  """

  def __init__(self,
               num_layers,
               num_units=512,
               num_heads=8,
               ffn_inner_dim=2048,
               dropout=0.1,
               attention_dropout=0.1,
               ffn_dropout=0.1,
               ffn_activation=tf.nn.relu,
               position_encoder_class=SinusoidalPositionEncoder,
               **kwargs):
    """Initializes the parameters of the encoder.

    Args:
      num_layers: The number of layers.
      num_units: The number of hidden units.
      num_heads: The number of heads in the multi-head attention.
      ffn_inner_dim: The number of units of the inner linear transformation
        in the feed forward layer.
      dropout: The probability to drop units from the outputs.
      attention_dropout: The probability to drop units from the attention.
      ffn_dropout: The probability to drop units from the activation output in
        the feed forward layer.
      ffn_activation: The activation function to apply between the two linear
        transformations of the feed forward layer.
      position_encoder_class: The :class:`opennmt.layers.PositionEncoder`
        class to use for position encoding (or a callable that returns an
        instance).
    """
    super(SelfAttentionEncoder, self).__init__(**kwargs)
    self.num_units = num_units
    self.dropout = dropout
    self.position_encoder = None
    if position_encoder_class is not None:
      self.position_encoder = position_encoder_class()
    self.layer_norm = common.LayerNorm()
    self.layers = [
        transformer.SelfAttentionEncoderLayer(
            num_units,
            num_heads,
            ffn_inner_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            ffn_dropout=ffn_dropout,
            ffn_activation=ffn_activation)
        for i in range(num_layers)]

  def call(self, inputs, sequence_length=None, training=None):
    inputs *= self.num_units**0.5
    if self.position_encoder is not None:
      inputs = self.position_encoder(inputs)
    inputs = common.dropout(inputs, self.dropout, training=training)
    mask = self.build_mask(inputs, sequence_length=sequence_length)
    for layer in self.layers:
      inputs = layer(inputs, mask=mask, training=training)
    outputs = self.layer_norm(inputs)
    return outputs, None, sequence_length

  def map_v1_weights(self, weights):
    m = []
    m += self.layer_norm.map_v1_weights(weights["LayerNorm"])
    for i, layer in enumerate(self.layers):
      m += layer.map_v1_weights(weights["layer_%d" % i])
    return m



"""Standalone script to detokenize a corpus."""

from __future__ import print_function

import argparse

from opennmt import tokenizers


def main():
  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument(
      "--delimiter", default=" ",
      help="Token delimiter used in text serialization.")
  parser.add_argument(
      "--tokenizer_config", default=None,
      help="Tokenization configuration.")
  args = parser.parse_args()

  tokenizer = tokenizers.make_tokenizer(args.config)
  tokenizer.detokenize_stream(delimiter=args.delimiter)

if __name__ == "__main__":
  main()



"""Define layers related to the Google's Transformer model."""

import tensorflow as tf
import numpy as np

from opennmt.layers import common
from opennmt.utils import misc


def _lower_triangle_mask(sequence_length, maximum_length=None, dtype=tf.bool):
  batch_size = tf.shape(sequence_length)[0]
  if maximum_length is None:
    maximum_length = tf.reduce_max(sequence_length)
  mask = tf.ones([batch_size, maximum_length, maximum_length], dtype=dtype)
  mask = tf.linalg.band_part(mask, -1, 0)
  return mask

def future_mask(sequence_length, maximum_length=None, dtype=tf.bool):
  """Builds the dot product mask for future positions.

  Args:
    sequence_length: The sequence length.
    maximum_length: Optional size of the returned time dimension. Otherwise
      it is the maximum of :obj:`sequence_length`.
    dtype: The type of the mask tensor.

  Returns:
    A ``tf.Tensor`` of type :obj:`dtype` and shape
    ``[batch_size, max_length, max_length]``.
  """
  sequence_mask = tf.sequence_mask(sequence_length, maxlen=maximum_length, dtype=dtype)
  sequence_mask = tf.expand_dims(sequence_mask, axis=1)
  mask = _lower_triangle_mask(sequence_length, maximum_length=maximum_length, dtype=dtype)
  if dtype is tf.bool:
    return tf.math.logical_and(mask, sequence_mask)
  else:
    return mask * sequence_mask

def split_heads(inputs, num_heads):
  """Splits a tensor in depth.

  Args:
    inputs: A ``tf.Tensor`` of shape :math:`[B, T, D]`.
    num_heads: The number of heads :math:`H`.

  Returns:
    A ``tf.Tensor`` of shape :math:`[B, H, T, D / H]`.
  """
  shape = misc.shape_list(inputs)
  outputs = tf.reshape(inputs, [shape[0], shape[1], num_heads, shape[2] // num_heads])
  outputs = tf.transpose(outputs, perm=[0, 2, 1, 3])
  return outputs

def combine_heads(inputs):
  """Concatenates heads.

  Args:
    inputs: A ``tf.Tensor`` of shape :math:`[B, H, T, D]`.

  Returns:
    A ``tf.Tensor`` of shape :math:`[B, T, D * H]`.
  """
  shape = misc.shape_list(inputs)
  outputs = tf.transpose(inputs, perm=[0, 2, 1, 3])
  outputs = tf.reshape(outputs, [shape[0], shape[2], shape[1] * shape[3]])
  return outputs


class FeedForwardNetwork(tf.keras.layers.Layer):
  """Implements the Transformer's "Feed Forward" layer.

  .. math::

      ffn(x) = max(0, x*W_1 + b_1)*W_2 + b_2
  """

  def __init__(self,
               inner_dim,
               output_dim,
               dropout=0.1,
               activation=tf.nn.relu,
               **kwargs):
    """Initializes this layer.

    Args:
      inner_dim: The number of units of the inner linear transformation.
      output_dim: The number of units of the ouput linear transformation.
      dropout: The probability to drop units from the activation output.
      activation: The activation function to apply between the two linear
        transformations.
      kwargs: Additional layer arguments.
    """
    super(FeedForwardNetwork, self).__init__(**kwargs)
    self.inner = common.Dense(inner_dim, activation=activation)
    self.outer = common.Dense(output_dim)
    self.dropout = dropout

  def call(self, inputs, training=None):  # pylint: disable=arguments-differ
    """Runs the layer."""
    inner = self.inner(inputs)
    inner = common.dropout(inner, self.dropout, training=training)
    return self.outer(inner)

  def map_v1_weights(self, weights):
    # V1 used conv1d layers that have a leading dimensions.
    weights = tf.nest.map_structure(np.squeeze, weights)
    m = []
    m += self.inner.map_v1_weights(weights["conv1d"])
    m += self.outer.map_v1_weights(weights["conv1d_1"])
    return m


class MultiHeadAttention(tf.keras.layers.Layer):
  """Computes the multi-head attention as described in
  https://arxiv.org/abs/1706.03762.
  """

  def __init__(self,
               num_heads,
               num_units,
               dropout=0.1,
               return_attention=False,
               **kwargs):
    """Initializes this layer.

    Args:
      num_heads: The number of attention heads.
      num_units: The number of hidden units.
      dropout: The probability to drop units from the inputs.
      return_attention: If ``True``, also return the attention weights of the
        first head.
      kwargs: Additional layer arguments.
    """
    super(MultiHeadAttention, self).__init__(**kwargs)
    if num_units % num_heads != 0:
      raise ValueError("Multi head attention requires that num_units is a"
                       " multiple of %s" % num_heads)
    self.num_heads = num_heads
    self.num_units = num_units
    self.linear_queries = common.Dense(num_units)
    self.linear_keys = common.Dense(num_units)
    self.linear_values = common.Dense(num_units)
    self.linear_output = common.Dense(num_units)
    self.dropout = dropout
    self.return_attention = return_attention

  def map_v1_weights(self, weights):
    # V1 used conv1d layers that have a leading dimensions.
    weights = tf.nest.map_structure(np.squeeze, weights)

    # V1 used fused linear projections, so the weights should be split accordingly.
    def _partial_weights(key, num_splits, index):
      return tf.nest.map_structure(
          lambda w: np.split(w, num_splits, axis=0 if w.ndim == 1 else 1)[index],
          weights[key])

    m = []
    if "conv1d_2" not in weights:  # Case self-attention.
      m += self.linear_queries.map_v1_weights(_partial_weights("conv1d", 3, 0))
      m += self.linear_keys.map_v1_weights(_partial_weights("conv1d", 3, 1))
      m += self.linear_values.map_v1_weights(_partial_weights("conv1d", 3, 2))
      m += self.linear_output.map_v1_weights(weights["conv1d_1"])
    else:
      m += self.linear_queries.map_v1_weights(weights["conv1d"])
      m += self.linear_keys.map_v1_weights(_partial_weights("conv1d_1", 2, 0))
      m += self.linear_values.map_v1_weights(_partial_weights("conv1d_1", 2, 1))
      m += self.linear_output.map_v1_weights(weights["conv1d_2"])
    return m

  def call(self, inputs, memory=None, mask=None, cache=None, training=None):  # pylint: disable=arguments-differ
    """Runs the layer.

    Args:
      inputs: The sequence of queries. A tensor of shape :math:`[B, T_1, ...]`.
      memory: The sequence to attend. A tensor of shape :math:`[B, T_2, ...]`.
        If ``None``, computes self-attention.
      mask: The dot product mask. A boolean tensor of shape :math:`[B, T_2]` or
        :math:`[B, T_1, T_2]`.
      cache: An optional tuple containing projected keys and values from the
        previous step. Tensors of shape :math:`[B, H, T_2, D / H]`.
      training: Run in training mode.

    Returns:
      A tuple with the attention context, the updated cache and the attention
      probabilities of the first head (if :obj:`return_attention` is ``True``).
    """

    def _compute_kv(x):
      keys = self.linear_keys(x)
      keys = split_heads(keys, self.num_heads)
      values = self.linear_values(x)
      values = split_heads(values, self.num_heads)
      return keys, values

    # Compute queries.
    queries = self.linear_queries(inputs)
    queries = split_heads(queries, self.num_heads)
    queries *= (self.num_units // self.num_heads)**-0.5

    # Compute keys and values.
    if memory is None:
      keys, values = _compute_kv(inputs)
      if cache:
        keys = tf.concat([cache[0], keys], axis=2)
        values = tf.concat([cache[1], values], axis=2)
    else:
      if cache:
        keys, values = tf.cond(
            tf.equal(tf.shape(cache[0])[2], 0),
            true_fn=lambda: _compute_kv(memory),
            false_fn=lambda: cache)
      else:
        keys, values = _compute_kv(memory)

    cache = (keys, values)

    # Dot product attention.
    dot = tf.matmul(queries, keys, transpose_b=True)
    if mask is not None:
      mask = tf.cast(mask, tf.float32)
      if mask.shape.rank == 2:
        mask = tf.expand_dims(mask, 1)  # Broadcast on time dimension.
      mask = tf.expand_dims(mask, 1)  # Broadcast on head dimension.
      dot = tf.cast(tf.cast(dot, tf.float32) * mask + ((1.0 - mask) * tf.float32.min), dot.dtype)
    attn = tf.cast(tf.nn.softmax(tf.cast(dot, tf.float32)), dot.dtype)
    drop_attn = common.dropout(attn, self.dropout, training=training)
    heads = tf.matmul(drop_attn, values)

    # Concatenate all heads output.
    combined = combine_heads(heads)
    outputs = self.linear_output(combined)
    if self.return_attention:
      return outputs, cache, attn
    return outputs, cache


class TransformerLayerWrapper(common.LayerWrapper):
  """Layer wrapper that applies a standard Transformer preprocessing and
  postprocessing:

  .. code-block:: text

      y = layer_norm(x)
      y = dropout(layer(y)) + x
  """

  def __init__(self, layer, output_dropout, **kwargs):
    """Initializes the wrapper.

    Args:
      layer: The Transformer layer to wrap.
      output_dropout: The dropout to apply on the layer output.
      **kwargs: Additional layer arguments.
    """
    super(TransformerLayerWrapper, self).__init__(
        layer,
        normalize_input=True,
        output_dropout=output_dropout,
        residual_connection=True,
        **kwargs)

  def map_v1_weights(self, weights):
    m = []
    m += self.input_layer_norm.map_v1_weights(weights["LayerNorm"])
    m += self.layer.map_v1_weights(weights)
    return m


class SelfAttentionEncoderLayer(tf.keras.layers.Layer):
  """Implements one self-attention encoding layer."""

  def __init__(self,
               num_units,
               num_heads,
               ffn_inner_dim,
               dropout=0.1,
               attention_dropout=0.1,
               ffn_dropout=0.1,
               ffn_activation=tf.nn.relu,
               **kwargs):
    """Initializes the layer.

    Args:
      num_units: The number of hidden units.
      num_heads: The number of heads in the multi-head attention.
      ffn_inner_dim: The number of units of the inner linear transformation
        in the feed forward layer.
      dropout: The probability to drop units from the outputs.
      attention_dropout: The probability to drop units from the attention.
      ffn_dropout: The probability to drop units from the activation output in
        the feed forward layer.
      ffn_activation: The activation function to apply between the two linear
        transformations of the feed forward layer.
      kwargs: Additional layer arguments.
    """
    super(SelfAttentionEncoderLayer, self).__init__(**kwargs)
    self.self_attention = MultiHeadAttention(
        num_heads, num_units, dropout=attention_dropout)
    self.self_attention = TransformerLayerWrapper(
        self.self_attention, dropout)
    self.ffn = FeedForwardNetwork(
        ffn_inner_dim,
        num_units,
        dropout=ffn_dropout,
        activation=ffn_activation)
    self.ffn = TransformerLayerWrapper(
        self.ffn, dropout)

  def call(self, x, mask=None, training=None):  # pylint: disable=arguments-differ
    """Runs the encoder layer."""
    y, _ = self.self_attention(x, mask=mask, training=training)
    y = self.ffn(y, training=training)
    return y

  def map_v1_weights(self, weights):
    m = []
    m += self.self_attention.map_v1_weights(weights["multi_head"])
    m += self.ffn.map_v1_weights(weights["ffn"])
    return m


class SelfAttentionDecoderLayer(tf.keras.layers.Layer):
  """Implements one self-attention decoding layer."""

  def __init__(self,
               num_units,
               num_heads,
               ffn_inner_dim,
               num_sources=1,
               dropout=0.1,
               attention_dropout=0.1,
               ffn_dropout=0.1,
               ffn_activation=tf.nn.relu,
               **kwargs):
    """Initializes the layer.

    Args:
      num_units: The number of hidden units.
      num_heads: The number of heads in the multi-head attention.
      ffn_inner_dim: The number of units of the inner linear transformation
        in the feed forward layer.
      num_sources: The number of source contexts.
      dropout: The probability to drop units from the outputs.
      attention_dropout: The probability to drop units from the attention.
      ffn_dropout: The probability to drop units from the activation output in
        the feed forward layer.
      ffn_activation: The activation function to apply between the two linear
        transformations of the feed forward layer.
      **kwargs: Additional layer arguments.
    """
    super(SelfAttentionDecoderLayer, self).__init__(**kwargs)
    self.self_attention = MultiHeadAttention(
        num_heads,
        num_units,
        dropout=attention_dropout)
    self.self_attention = TransformerLayerWrapper(
        self.self_attention, dropout)
    self.attention = []
    for _ in range(num_sources):
      attention = MultiHeadAttention(
          num_heads,
          num_units,
          dropout=attention_dropout,
          return_attention=num_sources == 1)
      attention = TransformerLayerWrapper(
          attention, dropout)
      self.attention.append(attention)
    self.ffn = FeedForwardNetwork(
        ffn_inner_dim,
        num_units,
        dropout=ffn_dropout,
        activation=ffn_activation)
    self.ffn = TransformerLayerWrapper(
        self.ffn, dropout)

  def map_v1_weights(self, weights):
    m = []
    m += self.self_attention.map_v1_weights(weights["masked_multi_head"])
    m += self.attention[0].map_v1_weights(weights["multi_head"])
    m += self.ffn.map_v1_weights(weights["ffn"])
    return m

  # pylint: disable=arguments-differ
  def call(self,
           inputs,
           mask=None,
           memory=None,
           memory_mask=None,
           cache=None,
           training=None):
    """Runs the decoder layer."""
    if cache is None:
      cache = {}

    outputs, self_kv = self.self_attention(
        inputs,
        mask=mask,
        cache=cache.get("self_kv"),
        training=training)

    attention = None
    memory_kv = []
    if memory is not None:
      memory_cache = cache.get("memory_kv")
      if memory_cache is None:
        memory_cache = [None] * len(self.attention)
      for layer, mem, mem_mask, mem_cache in zip(
          self.attention, memory, memory_mask, memory_cache):
        result = layer(
            outputs,
            memory=mem,
            mask=mem_mask,
            cache=mem_cache,
            training=training)
        if len(result) == 3:
          outputs, memory_kv_i, attention = result
          attention = attention[:, 0]  # Use the first head for the attention vector.
        else:
          outputs, memory_kv_i = result
        memory_kv.append(memory_kv_i)

    outputs = self.ffn(outputs, training=training)
    cache = dict(self_kv=self_kv, memory_kv=memory_kv)
    return outputs, cache, attention



"""Data manipulation module."""

from opennmt.data.dataset import batch_dataset
from opennmt.data.dataset import batch_sequence_dataset
from opennmt.data.dataset import filter_examples_by_length
from opennmt.data.dataset import filter_irregular_batches
from opennmt.data.dataset import get_dataset_size
from opennmt.data.dataset import inference_pipeline
from opennmt.data.dataset import random_shard
from opennmt.data.dataset import shuffle_dataset
from opennmt.data.dataset import training_pipeline

from opennmt.data.noise import Noise
from opennmt.data.noise import WordDropout
from opennmt.data.noise import WordNoiser
from opennmt.data.noise import WordOmission
from opennmt.data.noise import WordPermutation
from opennmt.data.noise import WordReplacement

from opennmt.data.text import alignment_matrix_from_pharaoh
from opennmt.data.text import tokens_to_chars
from opennmt.data.text import tokens_to_words

from opennmt.data.vocab import Vocab



"""Define position encoder classes."""

import math
import abc
import six

import tensorflow as tf

from opennmt.layers.reducer import SumReducer


@six.add_metaclass(abc.ABCMeta)
class PositionEncoder(tf.keras.layers.Layer):
  """Base class for position encoders."""

  def __init__(self, reducer=SumReducer(), **kwargs):
    """Initializes the position encoder.

    Args:
      reducer: A :class:`opennmt.layers.Reducer` to merge inputs and position
        encodings.
      **kwargs: Additional layer keyword arguments.
    """
    super(PositionEncoder, self).__init__(**kwargs)
    self.reducer = reducer

  def call(self, inputs, position=None):  # pylint: disable=arguments-differ
    """Add position encodings to :obj:`inputs`.

    Args:
      inputs: The inputs to encode.
      position: The single position to encode, to use when this layer is called
        step by step.

    Returns:
      A ``tf.Tensor`` whose shape depends on the configured ``reducer``.
    """
    batch_size = tf.shape(inputs)[0]
    timesteps = tf.shape(inputs)[1]
    input_dim = inputs.shape[-1]
    positions = tf.range(timesteps) + 1 if position is None else [position]
    position_encoding = self._encode([positions], input_dim)
    position_encoding = tf.tile(position_encoding, [batch_size, 1, 1])
    return self.reducer([inputs, position_encoding])

  @abc.abstractmethod
  def _encode(self, positions, depth):
    """Creates position encodings.

    Args:
      positions: The positions to encode of shape :math:`[B, ...]`.
      depth: The encoding depth :math:`D`.

    Returns:
      A ``tf.Tensor`` of shape :math:`[B, ..., D]`.
    """
    raise NotImplementedError()


class PositionEmbedder(PositionEncoder):
  """Encodes position with a lookup table."""

  def __init__(self, maximum_position=128, reducer=SumReducer(), **kwargs):
    """Initializes the position encoder.

    Args:
      maximum_position: The maximum position to embed. Positions greater
        than this value will be set to :obj:`maximum_position`.
      reducer: A :class:`opennmt.layers.Reducer` to merge inputs and position
        encodings.
      **kwargs: Additional layer keyword arguments.
    """
    super(PositionEmbedder, self).__init__(reducer=reducer, **kwargs)
    self.maximum_position = maximum_position
    self.embedding = None

  def build(self, input_shape):
    shape = [self.maximum_position + 1, input_shape[-1]]
    self.embedding = self.add_weight("position_embedding", shape)
    super(PositionEmbedder, self).build(input_shape)

  def _encode(self, positions, depth):
    positions = tf.minimum(positions, self.maximum_position)
    return tf.nn.embedding_lookup(self.embedding, positions)


class SinusoidalPositionEncoder(PositionEncoder):
  """Encodes positions with sine waves as described in
  https://arxiv.org/abs/1706.03762.
  """

  def _encode(self, positions, depth):
    if depth % 2 != 0:
      raise ValueError("SinusoidalPositionEncoder expects the depth to be divisble "
                       "by 2 but got %d" % depth)

    batch_size = tf.shape(positions)[0]
    positions = tf.cast(positions, tf.float32)

    log_timescale_increment = math.log(10000) / (depth / 2 - 1)
    inv_timescales = tf.exp(tf.range(depth / 2, dtype=tf.float32) * -log_timescale_increment)
    inv_timescales = tf.reshape(tf.tile(inv_timescales, [batch_size]), [batch_size, -1])
    scaled_time = tf.expand_dims(positions, -1) * tf.expand_dims(inv_timescales, 1)
    encoding = tf.concat([tf.sin(scaled_time), tf.cos(scaled_time)], axis=2)
    return tf.cast(encoding, self.dtype)



"""OpenNMT module."""

__version__ = "2.1.1"

from opennmt.config import load_config
from opennmt.config import load_model

from opennmt.constants import END_OF_SENTENCE_ID
from opennmt.constants import END_OF_SENTENCE_TOKEN
from opennmt.constants import PADDING_ID
from opennmt.constants import PADDING_TOKEN
from opennmt.constants import START_OF_SENTENCE_ID
from opennmt.constants import START_OF_SENTENCE_TOKEN
from opennmt.constants import UNKNOWN_TOKEN

from opennmt.runner import Runner



"""Catalog of predefined models."""

import tensorflow as tf
import tensorflow_addons as tfa

from opennmt import decoders
from opennmt import encoders
from opennmt import inputters
from opennmt import layers
from opennmt.models import language_model
from opennmt.models import sequence_tagger
from opennmt.models import sequence_to_sequence
from opennmt.models import transformer
from opennmt.utils.misc import merge_dict


class ListenAttendSpell(sequence_to_sequence.SequenceToSequence):
  """Defines a model similar to the "Listen, Attend and Spell" model described
  in https://arxiv.org/abs/1508.01211.
  """
  def __init__(self):
    super(ListenAttendSpell, self).__init__(
        source_inputter=inputters.SequenceRecordInputter(input_depth=40),
        target_inputter=inputters.WordEmbedder(
            embedding_size=50),
        encoder=encoders.PyramidalRNNEncoder(
            num_layers=3,
            num_units=512,
            reduction_factor=2,
            cell_class=tf.keras.layers.LSTMCell,
            dropout=0.3),
        decoder=decoders.AttentionalRNNDecoder(
            num_layers=3,
            num_units=512,
            attention_mechanism_class=tfa.seq2seq.LuongMonotonicAttention,
            cell_class=tf.keras.layers.LSTMCell,
            dropout=0.3,
            residual_connections=False,
            first_layer_attention=True))

  def auto_config(self, num_replicas=1):
    config = super(ListenAttendSpell, self).auto_config(num_replicas=num_replicas)
    return merge_dict(config, {
        "params": {
            "optimizer": "SGD",
            "learning_rate": 0.2,
            "scheduled_sampling_type": "constant",
            "scheduled_sampling_read_probability": 0.9
        },
        "train": {
            "batch_size": 32,
            "length_bucket_width": 15,
            "maximum_features_length": 2450,
            "maximum_labels_length": 330
        }
    })

class _RNNBase(sequence_to_sequence.SequenceToSequence):
  """Base class for RNN based NMT models."""
  def __init__(self, *args, **kwargs):
    super(_RNNBase, self).__init__(*args, **kwargs)

  def auto_config(self, num_replicas=1):
    config = super(_RNNBase, self).auto_config(num_replicas=num_replicas)
    return merge_dict(config, {
        "params": {
            "optimizer": "Adam",
            "learning_rate": 0.0002
        },
        "train": {
            "batch_size": 64,
            "maximum_features_length": 80,
            "maximum_labels_length": 80
        }
    })

class LuongAttention(_RNNBase):
  """Defines a LSTM encoder-decoder model as described in https://arxiv.org/abs/1508.04025."""
  def __init__(self):
    super(LuongAttention, self).__init__(
        source_inputter=inputters.WordEmbedder(
            embedding_size=512),
        target_inputter=inputters.WordEmbedder(
            embedding_size=512),
        encoder=encoders.RNNEncoder(
            num_layers=4,
            num_units=1000,
            dropout=0.2,
            residual_connections=False,
            cell_class=tf.keras.layers.LSTMCell),
        decoder=decoders.AttentionalRNNDecoder(
            num_layers=4,
            num_units=1000,
            bridge_class=layers.CopyBridge,
            attention_mechanism_class=tfa.seq2seq.LuongAttention,
            cell_class=tf.keras.layers.LSTMCell,
            dropout=0.2,
            residual_connections=False))

class LstmCnnCrfTagger(sequence_tagger.SequenceTagger):
  """Defines a bidirectional LSTM-CNNs-CRF as described in https://arxiv.org/abs/1603.01354."""
  def __init__(self):
    # pylint: disable=bad-continuation
    super(LstmCnnCrfTagger, self).__init__(
        inputter=inputters.MixedInputter([
            inputters.WordEmbedder(
                embedding_size=100),
            inputters.CharConvEmbedder(
                embedding_size=30,
                num_outputs=30,
                kernel_size=3,
                stride=1,
                dropout=0.5)],
            dropout=0.5),
        encoder=encoders.RNNEncoder(
            num_layers=1,
            num_units=400,
            bidirectional=True,
            dropout=0.5,
            residual_connections=False,
            cell_class=tf.keras.layers.LSTMCell),
        crf_decoding=True)

  def auto_config(self, num_replicas=1):
    config = super(SeqTagger, self).auto_config(num_replicas=num_replicas)
    return merge_dict(config, {
        "params": {
            "optimizer": "Adam",
            "learning_rate": 0.001
        },
        "train": {
            "batch_size": 32
        }
    })

class Transformer(transformer.Transformer):
  """Defines a Transformer model as decribed in https://arxiv.org/abs/1706.03762."""
  def __init__(self):
    super(Transformer, self).__init__(
        source_inputter=inputters.WordEmbedder(
            embedding_size=512),
        target_inputter=inputters.WordEmbedder(
            embedding_size=512),
        num_layers=6,
        num_units=512,
        num_heads=8,
        ffn_inner_dim=2048,
        dropout=0.1,
        attention_dropout=0.1,
        ffn_dropout=0.1)

class TransformerBig(transformer.Transformer):
  """Defines a large Transformer model as decribed in https://arxiv.org/abs/1706.03762."""
  def __init__(self):
    super(TransformerBig, self).__init__(
        source_inputter=inputters.WordEmbedder(
            embedding_size=1024),
        target_inputter=inputters.WordEmbedder(
            embedding_size=1024),
        num_layers=6,
        num_units=1024,
        num_heads=16,
        ffn_inner_dim=4096,
        dropout=0.3,
        attention_dropout=0.1,
        ffn_dropout=0.1)

class GPT2Small(language_model.LanguageModel):
  """GPT-2 language model (small version) as described in:

  https://d4mucfpksywv.cloudfront.net/better-language-models/language-models.pdf
  """

  def __init__(self):
    super(GPT2Small, self).__init__(
        decoder=decoders.SelfAttentionDecoder(
            num_layers=12,
            num_units=768,
            num_heads=12,
            ffn_inner_dim=3072,
            ffn_activation=layers.gelu,
            position_encoder_class=lambda: layers.PositionEmbedder(maximum_position=1024),
            num_sources=0),
        embedding_size=768)

  def auto_config(self, num_replicas=1):
    config = super(GPT2Small, self).auto_config(num_replicas=num_replicas)
    return merge_dict(config, {
        "params": {
            "average_loss_in_time": True,
            "optimizer": "Adam",
            "learning_rate": 2.5e-4,
            "decay_type": "CosineAnnealing",
            "decay_params": {
                "max_step": 1000000,
                "warmup_steps": 2000,
            }
        },
        "train": {
            # Below options are from GPT-1.
            "batch_size": 64,
            "maximum_features_length": 512
        }
    })



"""Define convolution-based encoders."""

import tensorflow as tf

from opennmt.encoders.encoder import Encoder
from opennmt.layers import common
from opennmt.layers import position


class ConvEncoder(Encoder):
  """An encoder that applies a convolution over the input sequence
  as described in https://arxiv.org/abs/1611.02344.
  """

  def __init__(self,
               num_layers_a,
               num_layers_c,
               num_units,
               kernel_size=3,
               dropout=0.3,
               position_encoder_class=position.PositionEmbedder):
    """Initializes the parameters of the encoder.

    Args:
      num_layers_a: The number of layers in CNN-a.
      num_layers_c: The number of layers in CNN-c.
      num_units: The number of output filters.
      kernel_size: The kernel size.
      dropout: The probability to drop units from the inputs.
      position_encoder_class: The :class:`opennmt.layers.PositionEncoder`
        class to use for position encoding (or a callable that returns an
        instance).
    """
    super(ConvEncoder, self).__init__()
    self.dropout = dropout
    self.position_encoder = None
    if position_encoder_class is not None:
      self.position_encoder = position_encoder_class()
    self.cnn_a = [
        tf.keras.layers.Conv1D(num_units, kernel_size, padding="same")
        for _ in range(num_layers_a)]
    self.cnn_c = [
        tf.keras.layers.Conv1D(num_units, kernel_size, padding="same")
        for _ in range(num_layers_c)]

  def call(self, inputs, sequence_length=None, training=None):
    if self.position_encoder is not None:
      inputs = self.position_encoder(inputs)
    inputs = common.dropout(inputs, self.dropout, training=training)

    cnn_a = _cnn_stack(self.cnn_a, inputs)
    cnn_c = _cnn_stack(self.cnn_c, inputs)

    outputs = cnn_a
    state = tf.reduce_mean(cnn_c, axis=1)
    return (outputs, state, sequence_length)

def _cnn_stack(layers, inputs):
  next_input = inputs

  for l, layer in enumerate(layers):
    outputs = layer(next_input)
    # Add residual connections past the first layer.
    if l > 0:
      outputs += next_input
    next_input = tf.tanh(outputs)

  return next_input



"""Define self-attention decoder."""

import tensorflow as tf

from opennmt.decoders import decoder
from opennmt.layers import common, transformer
from opennmt.layers.position import SinusoidalPositionEncoder


class SelfAttentionDecoder(decoder.Decoder):
  """Encoder using self-attention as described in
  https://arxiv.org/abs/1706.03762.
  """

  def __init__(self,
               num_layers,
               num_units=512,
               num_heads=8,
               ffn_inner_dim=2048,
               dropout=0.1,
               attention_dropout=0.1,
               ffn_dropout=0.1,
               ffn_activation=tf.nn.relu,
               position_encoder_class=SinusoidalPositionEncoder,
               num_sources=1,
               **kwargs):
    """Initializes the parameters of the decoder.

    Args:
      num_layers: The number of layers.
      num_units: The number of hidden units.
      num_heads: The number of heads in the multi-head attention.
      ffn_inner_dim: The number of units of the inner linear transformation
        in the feed forward layer.
      dropout: The probability to drop units from the outputs.
      attention_dropout: The probability to drop units from the attention.
      ffn_dropout: The probability to drop units from the activation output in
        the feed forward layer.
      ffn_activation: The activation function to apply between the two linear
        transformations of the feed forward layer.
      position_encoder_class: The :class:`opennmt.layers.PositionEncoder`
        class to use for position encoding (or a callable that returns an
        instance).
      num_sources: The number of source contexts expected by this decoder.
      **kwargs: Additional layer arguments.
    """
    super(SelfAttentionDecoder, self).__init__(num_sources=num_sources, **kwargs)
    self.num_units = num_units
    self.num_heads = num_heads
    self.dropout = dropout
    self.position_encoder = None
    if position_encoder_class is not None:
      self.position_encoder = position_encoder_class()
    self.layer_norm = common.LayerNorm()
    self.layers = [
        transformer.SelfAttentionDecoderLayer(
            self.num_units,
            self.num_heads,
            ffn_inner_dim,
            num_sources=num_sources,
            dropout=dropout,
            attention_dropout=attention_dropout,
            ffn_dropout=ffn_dropout,
            ffn_activation=ffn_activation)
        for i in range(num_layers)]

  @property
  def minimum_sources(self):
    return 0

  @property
  def maximum_sources(self):
    return 1e6  # An arbitrary large number.

  @property
  def support_alignment_history(self):
    return self.num_sources == 1

  def map_v1_weights(self, weights):
    m = []
    m += self.output_layer.map_v1_weights(weights["dense"])
    m += self.layer_norm.map_v1_weights(weights["LayerNorm"])
    for i, layer in enumerate(self.layers):
      m += layer.map_v1_weights(weights["layer_%d" % i])
    return m

  def _run(self,
           inputs,
           sequence_length=None,
           cache=None,
           memory=None,
           memory_sequence_length=None,
           step=None,
           training=None):
    # Process inputs.
    inputs *= self.num_units**0.5
    if self.position_encoder is not None:
      inputs = self.position_encoder(inputs, position=step + 1 if step is not None else None)
    inputs = common.dropout(inputs, self.dropout, training=training)

    # Prepare query mask.
    mask = None
    if step is None:
      maximum_length = tf.shape(inputs)[1]
      if sequence_length is None:
        batch_size = tf.shape(inputs)[0]
        sequence_length = tf.fill([batch_size], maximum_length)
      mask = transformer.future_mask(sequence_length, maximum_length=maximum_length)

    # Prepare memory mask.
    memory_mask = None
    if memory is not None:
      if not isinstance(memory, (list, tuple)):
        memory = (memory,)
    if memory_sequence_length is not None:
      if not isinstance(memory_sequence_length, (list, tuple)):
        memory_sequence_length = (memory_sequence_length,)
      memory_mask = [
          tf.sequence_mask(mem_length, maxlen=tf.shape(mem)[1])
          for mem, mem_length in zip(memory, memory_sequence_length)]

    # Run each layer.
    new_cache = []
    for i, layer in enumerate(self.layers):
      inputs, layer_cache, attention = layer(
          inputs,
          mask=mask,
          memory=memory,
          memory_mask=memory_mask,
          cache=cache[i] if cache is not None else None,
          training=training)
      new_cache.append(layer_cache)
    outputs = self.layer_norm(inputs)
    return outputs, new_cache, attention

  def forward(self,
              inputs,
              sequence_length=None,
              initial_state=None,
              memory=None,
              memory_sequence_length=None,
              input_fn=None,
              sampling_probability=None,
              training=None):
    _ = initial_state
    _ = input_fn
    if sampling_probability is not None:
      raise ValueError("Scheduled sampling is not supported by this decoder")
    outputs, state, attention = self._run(
        inputs,
        sequence_length=sequence_length,
        memory=memory,
        memory_sequence_length=memory_sequence_length,
        training=training)
    logits = self.output_layer(outputs)
    return logits, state, attention

  def step(self,
           inputs,
           timestep,
           state=None,
           memory=None,
           memory_sequence_length=None,
           training=None):
    inputs = tf.expand_dims(inputs, 1)
    outputs, state, attention = self._run(
        inputs,
        cache=state,
        memory=memory,
        memory_sequence_length=memory_sequence_length,
        step=timestep,
        training=training)
    outputs = tf.squeeze(outputs, axis=1)
    if attention is not None:
      attention = tf.squeeze(attention, axis=1)
    return outputs, state, attention

  def _get_initial_state(self, batch_size, dtype, initial_state=None):
    # The decoder state contains the keys and values projections of the previous timesteps.
    _ = initial_state
    cache = []
    for _ in self.layers:
      shape = [batch_size, self.num_heads, 0, self.num_units // self.num_heads]
      self_kv = (tf.zeros(shape, dtype=dtype), tf.zeros(shape, dtype=dtype))
      memory_kv = [
          (tf.zeros(shape, dtype=dtype), tf.zeros(shape, dtype=dtype))
          for _ in range(self.num_sources)]
      cache.append(dict(self_kv=self_kv, memory_kv=memory_kv))
    return cache



"""Hypotheses file scoring."""

import abc
import six

import tensorflow as tf


@six.add_metaclass(abc.ABCMeta)
class Scorer(object):
  """Scores hypotheses against references."""

  def __init__(self, name):
    self._name = name

  @property
  def name(self):
    """The scorer name."""
    return self._name

  @property
  def scores_name(self):
    """The names of returned scores."""
    return {self._name}

  @abc.abstractmethod
  def __call__(self, ref_path, hyp_path):
    """Scores hypotheses.

    Args:
      ref_path: Path to the reference.
      hyp_path: Path to the hypotheses.

    Returns:
      The score or dictionary of scores.
    """
    raise NotImplementedError()

  def lower_is_better(self):
    """Returns ``True`` if a lower score is better."""
    return False

  def higher_is_better(self):
    """Returns ``True`` if a higher score is better."""
    return not self.lower_is_better()


class ROUGEScorer(Scorer):
  """ROUGE scorer based on https://github.com/pltrdy/rouge."""

  def __init__(self):
    super(ROUGEScorer, self).__init__("rouge")

  @property
  def scores_name(self):
    return {"rouge-1", "rouge-2", "rouge-l"}

  def __call__(self, ref_path, hyp_path):
    from rouge import FilesRouge
    files_rouge = FilesRouge(hyp_path, ref_path)
    rouge_scores = files_rouge.get_scores(avg=True)
    return {name:rouge_scores[name]["f"] for name in self.scores_name}


class BLEUScorer(Scorer):
  """Scorer using sacreBLEU."""

  def __init__(self):
    try:
      import sacrebleu  # pylint: disable=unused-import, unused-variable
    except ImportError:
      raise ImportError("sacreBLEU evaluator requires Python 3")
    super(BLEUScorer, self).__init__("bleu")

  def __call__(self, ref_path, hyp_path):
    from sacrebleu import corpus_bleu
    with tf.io.gfile.GFile(ref_path) as ref_stream, tf.io.gfile.GFile(hyp_path) as sys_stream:
      bleu = corpus_bleu(sys_stream, [ref_stream])
      return bleu.score


def make_scorers(names):
  """Returns a list of scorers.

  Args:
    names: A list of scorer names or a single name.

  Returns:
    A list of :class:`opennmt.evaluation.Scorer` instances.

  Raises:
    ValueError: if a scorer name is invalid.
  """
  if not isinstance(names, list):
    names = [names]
  scorers = []
  for name in names:
    name = name.lower()
    scorer = None
    if name == "bleu":
      scorer = BLEUScorer()
    elif name == "rouge":
      scorer = ROUGEScorer()
    else:
      raise ValueError("No scorer associated with the name: {}".format(name))
    scorers.append(scorer)
  return scorers



from setuptools import setup, find_packages

tests_require = [
    "parameterized",
    "nose2"
]

setup(
    name="OpenNMT-tf",
    version="2.1.1",
    license="MIT",
    description="Neural machine translation and sequence learning using TensorFlow",
    author="OpenNMT",
    author_email="guillaume.klein@systrangroup.com",
    url="http://opennmt.net",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
    project_urls={
        "Documentation": "http://opennmt.net/OpenNMT-tf/",
        "Forum": "http://forum.opennmt.net/",
        "Gitter": "https://gitter.im/OpenNMT/OpenNMT-tf",
        "Source": "https://github.com/OpenNMT/OpenNMT-tf/"
    },
    keywords="tensorflow opennmt nmt neural machine translation",
    install_requires=[
        "pyonmttok>=1.16.0,<2;platform_system=='Linux'",
        "pyyaml==5.1.*",
        "rouge==0.3.1",
        "sacrebleu>=1.4.1,<2;python_version>='3.0'",
        "tensorflow-addons>=0.6.0"
    ],
    extras_require={
        "tests": tests_require,
    },
    tests_require=tests_require,
    test_suite="nose2.collector.collector",
    packages=find_packages(exclude=["bin", "*.tests"]),
    entry_points={
        "console_scripts": [
            "onmt-ark-to-records=opennmt.bin.ark_to_records:main",
            "onmt-build-vocab=opennmt.bin.build_vocab:main",
            "onmt-detokenize-text=opennmt.bin.detokenize_text:main",
            "onmt-main=opennmt.bin.main:main",
            "onmt-merge-config=opennmt.bin.merge_config:main",
            "onmt-tokenize-text=opennmt.bin.tokenize_text:main",
        ],
    }
)



"""Define inputters reading from TFRecord files."""

import tensorflow as tf
import numpy as np

from opennmt.inputters.inputter import Inputter
from opennmt.utils import misc


class SequenceRecordInputter(Inputter):
  """Inputter that reads ``tf.train.SequenceExample``.

  See Also:
    :func:`opennmt.inputters.create_sequence_records` to generate a compatible
    dataset.
  """

  def __init__(self, input_depth, **kwargs):
    """Initializes the parameters of the record inputter.

    Args:
      input_depth: The depth dimension of the input vectors.
      **kwargs: Additional layer keyword arguments.
    """
    super(SequenceRecordInputter, self).__init__(**kwargs)
    self.input_depth = input_depth

  def make_dataset(self, data_file, training=None):
    return tf.data.TFRecordDataset(
        data_file, compression_type="GZIP" if misc.is_gzip_file(data_file) else None)

  def input_signature(self):
    return {
        "tensor": tf.TensorSpec([None, None, self.input_depth], self.dtype),
        "length": tf.TensorSpec([None], tf.int32)
    }

  def make_features(self, element=None, features=None, training=None):
    if features is None:
      features = {}
    if "tensor" in features:
      return features
    _, feature_lists = tf.io.parse_single_sequence_example(element, sequence_features={
        "values": tf.io.FixedLenSequenceFeature([self.input_depth], dtype=tf.float32)})
    tensor = feature_lists["values"]
    features["length"] = tf.shape(tensor)[0]
    features["tensor"] = tf.cast(tensor, self.dtype)
    return features

  def call(self, features, training=None):
    return features["tensor"]


def write_sequence_record(vector, writer):
  """Writes a sequence vector as a TFRecord.

  Args:
    vector: A 2D Numpy float array of shape :math:`[T, D]`.
    writer: A ``tf.io.TFRecordWriter``.

  See Also:
    - :class:`opennmt.inputters.SequenceRecordInputter`
    - :func:`opennmt.inputters.create_sequence_records`
  """
  feature_list = tf.train.FeatureList(feature=[
      tf.train.Feature(float_list=tf.train.FloatList(value=values))
      for values in vector.astype(np.float32)])
  feature_lists = tf.train.FeatureLists(feature_list={"values": feature_list})
  example = tf.train.SequenceExample(feature_lists=feature_lists)
  writer.write(example.SerializeToString())

def create_sequence_records(vectors, path, compression=None):
  """Creates a TFRecord file of sequence vectors.

  Args:
    vectors: An iterable of 2D Numpy float arrays of shape :math:`[T, D]`.
    path: The output TFRecord file.
    compression: Optional compression type, can be "GZIP".

  Returns:
    Path to the TFRecord file. In most cases this is the same as :obj:`path` but
    if GZIP compression is enabled, the ".gz" extension is added if not already
    present.

  Raises:
    ValueError: if :obj:`compression` is invalid.

  See Also:
    - :class:`opennmt.inputters.SequenceRecordInputter`
    - :func:`opennmt.inputters.write_sequence_record`
  """
  if compression is not None:
    if compression not in ("GZIP",):
      raise ValueError("invalid compression type: %s" % compression)
    if compression == "GZIP" and not path.endswith(".gz"):
      path = "%s.gz" % path
  writer = tf.io.TFRecordWriter(path, options=compression)
  for vector in vectors:
    write_sequence_record(vector, writer)
  writer.close()
  return path
